"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import unittest
from math import isnan

import numpy as np

from hyram.phys import api, Fluid
from hyram.utilities import misc_utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

VERBOSE = False


class EtkTpdTestCase(unittest.TestCase):
    """
    Test engineering toolkit temperature, pressure, density API interface
    """
    def setUp(self):
        self.species = 'h2'
        self.phase = None
        self.phase_sat_vap = 'gas'  # saturated vapor
        self.phase_sat_liq = 'liquid'  # saturated vapor
        self.temperature = 298  # K
        self.pressure = 350e5  # Pa
        self.pressure_sat_vap = 1e6  # Pa, pressure for saturated phase
        self.density = 23  # kg/m3
        self.density_lh2 = 71  # kg/m3

    def test_compute_density_from_temperature_and_pressure(self):
        density, _ = api.compute_thermo_param(species=self.species,
                                              temp=self.temperature,
                                              pres=self.pressure)
        self.assertGreater(density, 0)

    def test_compute_density_from_phase_and_pressure(self):
        density, temp = api.compute_thermo_param(species=self.species,
                                                 phase=self.phase_sat_vap,
                                                 pres=self.pressure_sat_vap)
        self.assertGreater(density, 0)

    def test_compute_pressure_from_temperature_and_density(self):
        pressure, _ = api.compute_thermo_param(species=self.species,
                                               temp=self.temperature,
                                               density=self.density)
        self.assertGreater(pressure, 0)

    def test_compute_pressure_from_phase_and_density(self):
        pressure, temp = api.compute_thermo_param(species=self.species,
                                                  phase=self.phase_sat_vap,
                                                  density=self.density)
        self.assertGreater(pressure, 0)

    def test_compute_pressure_from_liq_phase_and_density(self):
        pressure, temp = api.compute_thermo_param(species=self.species,
                                                  phase=self.phase_sat_liq,
                                                  density=self.density_lh2)
        self.assertGreater(pressure, 0)

    def test_compute_temperature_from_pressure_and_density(self):
        temperature, _ = api.compute_thermo_param(species=self.species,
                                                  pres=self.pressure,
                                                  density=self.density)
        self.assertGreater(temperature, 0)

    def test_compute_temperature_from_pressure_and_phase(self):
        temperature, _ = api.compute_thermo_param(species=self.species,
                                                  phase=self.phase_sat_vap,
                                                  pres=self.pressure_sat_vap)
        self.assertGreater(temperature, 0)

    def test_reject_too_many_params(self):
        self.assertRaises(ValueError,
                          api.compute_thermo_param,
                          species=self.species,
                          phase=self.phase,
                          temp=self.temperature,
                          pres=self.pressure,
                          density=self.density)

    def test_reject_too_few_params(self):
        self.assertRaises(ValueError,
                          api.compute_thermo_param,
                          species=None,
                          phase=None,
                          temp=None,
                          pres=None,
                          density=None)


class TestETKTankMass(unittest.TestCase):
    """
    Test engineering toolkit tank mass API interface
    """
    def setUp(self):
        self.species = 'h2'
        self.phase = None
        self.phase_sat_vap = 'gas'  # saturated vapor
        self.phase_sat_liq = 'liquid'  # saturated vapor
        self.temperature = 298  # K
        self.pressure = 350e5  # Pa
        self.pressure_sat_vap = 1e6  # Pa, pressure for saturated phase
        self.volume = 1  # m3
        self.mass = 1  # kg
        self.mass_sat_liq = 70.85  # kg

    def test_calc_temp_from_pres_vol_mass(self):
        temp, _ = api.compute_tank_mass_param(species=self.species,
                                              pres=self.pressure,
                                              vol=self.volume,
                                              mass=self.mass)
        self.assertGreater(temp, 0)

    def test_calc_pres_from_temp_vol_mass(self):
        pres, _ = api.compute_tank_mass_param(species=self.species,
                                              temp=self.temperature,
                                              vol=self.volume,
                                              mass=self.mass)
        self.assertGreater(pres, 0)

    def test_calc_pres_temp_from_phase_vol_mass(self):
        pres, temp = api.compute_tank_mass_param(species=self.species,
                                                 phase=self.phase_sat_liq,
                                                 vol=self.volume,
                                                 mass=self.mass_sat_liq)
        self.assertGreater(pres, 0)
        self.assertGreater(temp, 0)

    def test_calc_vol_from_temp_pres_mass(self):
        vol, _ = api.compute_tank_mass_param(species=self.species,
                                             temp=self.temperature,
                                             pres=self.pressure,
                                             mass=self.mass)
        self.assertGreater(vol, 0)

    def test_calc_vol_temp_from_phase_pres_mass(self):
        vol, temp = api.compute_tank_mass_param(species=self.species,
                                                phase=self.phase_sat_liq,
                                                pres=self.pressure_sat_vap,
                                                mass=self.mass_sat_liq)
        self.assertGreater(vol, 0)
        self.assertGreater(temp, 0)

    def test_calc_mass_from_temp_pres_vol(self):
        mass, _ = api.compute_tank_mass_param(species=self.species,
                                              temp=self.temperature,
                                              pres=self.pressure,
                                              vol=self.volume)
        self.assertGreater(mass, 0)

    def test_calc_mass_temp_from_phase_pres_vol(self):
        mass, temp = api.compute_tank_mass_param(species=self.species,
                                                 phase=self.phase_sat_liq,
                                                 pres=self.pressure_sat_vap,
                                                 vol=self.volume)
        self.assertGreater(mass, 0)
        self.assertGreater(temp, 0)

    def test_reject_too_many_params(self):
        self.assertRaises(ValueError,
                          api.compute_tank_mass_param,
                          species=self.species,
                          phase=self.phase,
                          temp=self.temperature,
                          pres=self.pressure,
                          vol=self.volume,
                          mass=self.mass)

    def test_reject_too_few_params(self):
        self.assertRaises(ValueError,
                          api.compute_tank_mass_param,
                          species=None,
                          phase=None,
                          temp=None,
                          pres=None,
                          vol=None,
                          mass=None)


class TestETKTNTEquivalentMass(unittest.TestCase):
    """
    Test engineering toolkit TNT equivalence mass API interface
    """
    def test_calc_equiv_TNT_mass(self):
        flammable_mass = 1  # kg
        percent_yield = 3
        species = 'H2'
        equiv_TNT_mass = api.compute_equivalent_tnt_mass(vapor_mass=flammable_mass,
                                                         percent_yield=percent_yield,
                                                         species=species)
        # Hand calculation based on 1 kg H2 with TNT blast energy 4.68 MJ/kg and yield of 3%
        # self.assertAlmostEqual(equiv_TNT_mass, 0.03*1*1.20E+08/4.68e6)
        self.assertAlmostEqual(equiv_TNT_mass, 0.769, places=3)


class PlumeDispersionTestCase(unittest.TestCase):
    """
    Test plume dispersion physics API interface
    """
    def setUp(self):
        # Default inputs
        self.amb_fluid = Fluid(species='air', T=298, P=101325)
        self.rel_fluid = Fluid(species='hydrogen', T=298, P=35e6)
        self.mass_flow = None
        self.orif_diam = 3.56 / 1000  # m
        self.rel_angle = 0  # radians, 0=horizontal
        self.dis_coeff = 1
        self.nozzle_model = 'yuce'
        self.create_plot = True
        self.contour = 0.04  # mole fraction
        self.contour_min = 0  # mole fraction
        self.contour_max = 0.1  # mole fraction
        self.xmin = -2.5  # m
        self.xmax = 2.5  # m
        self.ymin = 0  # m
        self.ymax = 10  # m
        self.plot_title = "Mole Fraction of Leak"
        self.filename = None  # use default
        self.output_dir = None  # use default
        self.verbose = False

    def test_valid_input(self):
        result = api.analyze_jet_plume(self.amb_fluid, self.rel_fluid, self.orif_diam)
        self.assertTrue(result is not None)

    def test_blend(self):
        # self.rel_fluid = Fluid({'h2': 1, 'ch4': 0.}, T=298, P=35e6)
        self.rel_fluid = Fluid({'h2': 0.05, 'ch4': 0.95}, T=298, P=1e6)
        result = api.analyze_jet_plume(self.amb_fluid, self.rel_fluid, self.orif_diam)
        self.assertTrue(result is not None)

    def test_reject_mole_frac_0(self):
        contour = 0  # mole fraction
        self.assertRaises(ValueError, api.analyze_jet_plume,
                          self.amb_fluid, self.rel_fluid, self.orif_diam, self.mass_flow,
                          self.rel_angle, self.dis_coeff, self.nozzle_model,
                          self.create_plot, contour,
                          self.contour_min, self.contour_max,
                          self.xmin, self.xmax, self.ymin, self.ymax,
                          self.plot_title, self.filename, self.output_dir,
                          self.verbose)

    def test_reject_mole_frac_1(self):
        contour = 1  # mole fraction
        self.assertRaises(ValueError, api.analyze_jet_plume,
                          self.amb_fluid, self.rel_fluid, self.orif_diam, self.mass_flow,
                          self.rel_angle, self.dis_coeff, self.nozzle_model,
                          self.create_plot, contour,
                          self.contour_min, self.contour_max,
                          self.xmin, self.xmax, self.ymin, self.ymax,
                          self.plot_title, self.filename, self.output_dir,
                          self.verbose)

    def test_reject_xmin_greater_than_xmax(self):
        xmin = 2
        xmax = 0
        self.assertRaises(ValueError, api.analyze_jet_plume,
                          self.amb_fluid, self.rel_fluid, self.orif_diam, self.mass_flow,
                          self.rel_angle, self.dis_coeff, self.nozzle_model,
                          self.create_plot, self.contour,
                          self.contour_min, self.contour_max,
                          xmin, xmax, self.ymin, self.ymax,
                          self.plot_title, self.filename, self.output_dir,
                          self.verbose)

    def test_reject_ymin_greater_than_ymax(self):
        ymin = 2
        ymax = 0
        self.assertRaises(ValueError, api.analyze_jet_plume,
                          self.amb_fluid, self.rel_fluid, self.orif_diam, self.mass_flow,
                          self.rel_angle, self.dis_coeff, self.nozzle_model,
                          self.create_plot, self.contour,
                          self.contour_min, self.contour_max,
                          self.xmin, self.xmax, ymin, ymax,
                          self.plot_title, self.filename, self.output_dir,
                          self.verbose)

    def test_default_inputs(self):
        results = api.analyze_jet_plume(self.amb_fluid,
                                        self.rel_fluid,
                                        self.orif_diam,
                                        self.mass_flow,
                                        self.rel_angle,
                                        self.dis_coeff,
                                        self.nozzle_model,
                                        self.create_plot,
                                        self.contour,
                                        self.contour_min,
                                        self.contour_max,
                                        self.xmin,
                                        self.xmax,
                                        self.ymin,
                                        self.ymax,
                                        self.plot_title,
                                        self.filename,
                                        self.output_dir,
                                        self.verbose)
        self.assertTrue(os.path.isfile(results['plot']))
        self.assertGreater(results['mass_flow_rate'], 0)
        self.assertGreater(results['streamline_dists'], 0)
        self.assertGreaterEqual(results['mole_frac_dists'][self.contour][0][1], 0)


class AccumulationTestCase(unittest.TestCase):
    """
    Test phys API accumulation (indoor release).
    """
    def setUp(self):
        # Default inputs
        output_dir = misc_utils.get_temp_folder()
        amb_fluid = Fluid(species='air', T=288, P=101325)
        rel_fluid = Fluid(species='hydrogen', T=287.8, P=35e6)

        self.times = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                               20, 21, 22, 23, 24, 25,
                               26, 27, 28, 29, 29.5])

        self.params = {
            'amb_fluid': amb_fluid,
            'rel_fluid': rel_fluid,
            'ceil_vent_xarea': 0.09079,
            'ceil_vent_height': 2.42,
            'dist_rel_to_wall': 2.1255,
            'enclos_height': 2.72,
            'floor_ceil_area': 16.72216,
            'floor_vent_height': 0.044,
            'floor_vent_xarea': 0.00762,
            'is_steady': False,
            'nozzle_key': 'yuce',
            'orif_diam': 3.56 / 1000,
            'orif_dis_coeff': 1,

            'rel_angle': 0,
            'rel_height': 0,
            'tank_volume': 0.00363,
            # 'ceil_vent_coeff': 1,
            # 'floor_vent_coeff': 1,
            'vol_flow_rate': 0,
            'times': self.times,
            'tmax': 30,
            'create_plots': True,

            'output_dir': output_dir,
            'verbose': VERBOSE
        }

    def test_defaults(self):
        num_times = len(self.times)
        rdict = api.analyze_accumulation(**self.params)

        self.assertTrue(rdict['status'])

        self.assertTrue('pressures_per_time' in rdict)
        pressures = rdict['pressures_per_time']
        self.assertEqual(len(pressures), num_times)

        self.assertTrue('depths' in rdict)
        depths = rdict['depths']
        self.assertEqual(len(depths), num_times)

        self.assertTrue('concentrations' in rdict)
        concs = rdict['concentrations']
        self.assertEqual(len(concs), num_times)

        self.assertTrue('overpressure' in rdict)
        overp = rdict['overpressure']
        self.assertEqual(type(overp), np.float64)

        self.assertTrue('time_of_overp' in rdict)
        t_overp = rdict['time_of_overp']
        self.assertEqual(type(t_overp), np.float64)

        self.assertTrue('pres_plot_filepath' in rdict)
        plot = rdict['pres_plot_filepath']
        self.assertEqual(type(plot), str)
        self.assertTrue(os.path.isfile(plot))

        self.assertTrue('mass_plot_filepath' in rdict)
        plot = rdict['mass_plot_filepath']
        self.assertEqual(type(plot), str)
        self.assertTrue(os.path.isfile(plot))

        self.assertTrue('layer_plot_filepath' in rdict)
        plot = rdict['layer_plot_filepath']
        self.assertEqual(type(plot), str)
        self.assertTrue(os.path.isfile(plot))

        self.assertTrue('trajectory_plot_filepath' in rdict)
        plot = rdict['trajectory_plot_filepath']
        self.assertEqual(type(plot), str)
        self.assertTrue(os.path.isfile(plot))

        self.assertTrue('mass_flow_plot_filepath' in rdict)
        plot = rdict['mass_plot_filepath']
        self.assertEqual(type(plot), str)
        self.assertTrue(os.path.isfile(plot))


class JetFlameAnalysisTestCase(unittest.TestCase):
    """
    Test jet flame analysis physics API interface
    """
    def test_default_jet_flame_analysis(self):
        amb_fluid = api.create_fluid(species='air', temp=298, pres=101325)
        rel_fluid = api.create_fluid(species='h2', temp=298, pres=35e6)
        orif_diam = 3.56 / 1000  # m
        dis_coeff = 1
        rel_angle = 0
        mass_flow = None
        nozzle_key = 'yuce'
        rel_humid = 0.89
        create_temp_plot = True
        temp_plot_filename = None
        txmin = 0
        txmax = 10
        tymin = -3.5
        tymax = 3.5

        analyze_flux = True
        heatflux_plot_filename = None
        flux_coordinates = [(0.01, 1, 0.01),
                            (0.5, 1, 0.5),
                            (1, 1, 0.5),
                            (2, 1, 1),
                            (2.5, 1, 1),
                            (5, 2, 1),
                            (10, 2, 0.5),
                            (15, 2, 0.5),
                            (25, 2, 1),
                            (40, 2, 2)]
        fxmin = -5
        fxmax = 15
        fymin = -10
        fymax = 10
        fzmin = -10
        fzmax = 10

        output_dir = misc_utils.get_temp_folder()
        verbose = False

        (temp_plot_filepath, heatflux_filepath,
         pos_flux, mass_flow, srad, visible_length, radiant_frac
         ) = api.jet_flame_analysis(amb_fluid, rel_fluid, orif_diam, mass_flow=None, dis_coeff=dis_coeff,
                                    rel_angle=rel_angle, nozzle_key=nozzle_key, rel_humid=rel_humid,
                                    create_temp_plot=create_temp_plot, temp_plot_filename=temp_plot_filename,
                                    temp_contours=None,
                                    temp_xlims=(txmin, txmax), temp_ylims=(tymin, tymax),

                                    analyze_flux=analyze_flux, flux_plot_filename=heatflux_plot_filename,
                                    flux_coordinates=flux_coordinates, flux_contours=None,
                                    flux_xlims=(fxmin, fxmax), flux_ylims=(fymin, fymax), flux_zlims=(fzmin, fzmax),

                                    output_dir=output_dir, verbose=verbose)
        self.assertTrue(os.path.exists(temp_plot_filepath))
        self.assertTrue(os.path.exists(heatflux_filepath))
        for flux_value in pos_flux:
            self.assertGreater(flux_value, 0)
        self.assertGreater(mass_flow, 0)
        self.assertGreater(srad, 0)
        self.assertGreater(visible_length, 0)
        self.assertGreater(radiant_frac, 0)


class OverpressureTestCase(unittest.TestCase):
    """
    Test overpressure calculation
    """
    def setUp(self):
        self.output_dir = misc_utils.get_temp_folder()

        # Default values
        self.locations = [(1, 1, 0)]  # distance is sqrt(2) m
        P_inf = 101325.  # Pa, ambient pressure
        jet_T = 300  # K, jet exit temperature
        jet_P = 60e5  # Pa, jet exit pressure
        self.ambient_fluid = Fluid(species='air', P = P_inf, T = jet_T)
        self.release_fluid = Fluid(species = 'hydrogen', T = jet_T, P = jet_P)
        self.orifice_diameter = 2*0.0254  # m, jet exit diameter

    def test_basic_bst_overpressure(self):
        method = 'BST'
        BST_mach_flame_speed = 0.35
        result_dict = api.compute_overpressure(method, self.locations,
                                               ambient_fluid=self.ambient_fluid,
                                               release_fluid=self.release_fluid,
                                               orifice_diameter=self.orifice_diameter,
                                               bst_flame_speed=BST_mach_flame_speed,
                                               output_dir=self.output_dir)
        overpressure = result_dict["overpressures"]
        impulse = result_dict["impulses"]
        overpressure_figure_path = result_dict["overp_plot_filepath"]
        impulse_figure_path = result_dict["impulse_plot_filepath"]
        self.assertTrue(result_dict is not None)
        self.assertGreaterEqual(overpressure, 0)
        self.assertGreaterEqual(impulse, 0)
        self.assertTrue(os.path.exists(overpressure_figure_path))
        self.assertTrue(os.path.exists(impulse_figure_path))

    def test_basic_tnt_overpressure(self):
        method = 'TNT'
        tnt_factor = 0.03
        result_dict = api.compute_overpressure(method, self.locations,
                                               ambient_fluid=self.ambient_fluid,
                                               release_fluid=self.release_fluid,
                                               orifice_diameter=self.orifice_diameter,
                                               tnt_factor=tnt_factor,
                                               output_dir=self.output_dir)
        overpressure = result_dict["overpressures"]
        impulse = result_dict["impulses"]
        overpressure_figure_path = result_dict["overp_plot_filepath"]
        impulse_figure_path = result_dict["impulse_plot_filepath"]
        self.assertTrue(result_dict is not None)
        self.assertGreaterEqual(overpressure, 0)
        self.assertGreaterEqual(impulse, 0)
        self.assertTrue(os.path.exists(overpressure_figure_path))
        self.assertTrue(os.path.exists(impulse_figure_path))

    def test_basic_bauwens_overpressure(self):
        method = 'Bauwens'
        result_dict = api.compute_overpressure(method, self.locations,
                                               ambient_fluid=self.ambient_fluid,
                                               release_fluid=self.release_fluid,
                                               orifice_diameter=self.orifice_diameter,
                                               output_dir=self.output_dir)
        overpressure = result_dict["overpressures"]
        impulse = result_dict["impulses"]
        overpressure_figure_path = result_dict["overp_plot_filepath"]
        impulse_figure_path = result_dict["impulse_plot_filepath"]
        self.assertTrue(result_dict is not None)
        self.assertGreaterEqual(overpressure, 0)
        self.assertTrue(isnan(impulse[0]))
        self.assertTrue(os.path.exists(overpressure_figure_path))
        self.assertIsNone(impulse_figure_path)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import numpy as np

from . import probits


def thermal_consequence(physical_responses,
                        consequence_modeling_decisions):
    """
    Calculates thermal fatality probabilities for all positions and leak sizes

    Parameters
    ----------
    physical_responses : dict
        Dictionary holding all physical responses associated with event

        qrads : array (item in physical_responses dict)
            Heat flux data (W/m2) for all positions and leak sizes
            1-D array, all leak sizes for location 1,
            then all leak sizes location 2, etc.

    consequence_modeling_decisions : dict
        Dictionary holding modeling decisions/specifications used to enacted the desired consequence model

        probit_thermal_id : str (item in consequence_modeling_decisions dict)
            String to identify thermal probit to use
            See probits.py for details

        exposure_time : float (item in consequence_modeling_decisions dict)
            Exposure time (seconds) for use in thermal probit
            
    Returns
    -------
    thermal_fatality_probs : array
        Probability of fatality from thermal effects
    """
    thermal_fatality_probs = [] 
    for qrad in physical_responses['qrads']:
        p_therm_fatal = probits.compute_thermal_fatality_prob(consequence_modeling_decisions['probit_thermal_id'],
                                                              qrad,
                                                              consequence_modeling_decisions['exposure_time'])
        thermal_fatality_probs.append(p_therm_fatal)

    return thermal_fatality_probs

def overpressure_consequence(physical_responses,
                             consequence_modeling_decisions):
    """
    Calculates overpressure fatality probabilities
    for all positions and leak sizes

    Parameters
    ----------
    physical_responses : dict
        Dictionary holding all physical responses associated with event

        overpressures : array (item in physical_responses dict)
            Peak overpressure values (Pa) for all positions and leak sizes
            1-D array, all leak sizes for location 1,
            then all leak sizes location 2, etc.

        impulses : array (item in physical_responses dict)
            Impulse values (Pa*s) for all positions and leak sizes
            1-D array, all leak sizes for location 1,
            then all leak sizes location 2, etc.

    consequence_modeling_decisions : dict
        Dictionary holding modeling decisions/specifications used to enacted the desired consequence model

        overpressure_probit_id : str (item in consequence_modeling_decisions dict)
            String to identify thermal probit to use
            See probits.py for details

    Returns
    -------
    overp_fatality_probs : array
        Probability of fatality from overpressure effects
    """
    overp_fatality_probs = []
    for overpressure, impulse in zip(physical_responses['overpressures'], physical_responses['impulses']):
        p_overp_fatal = probits.compute_overpressure_fatality_prob(consequence_modeling_decisions['probit_overp_id'],
                                                                   overpressure,
                                                                   impulse)
        overp_fatality_probs.append(p_overp_fatal)

    return overp_fatality_probs

def convert_consequence_to_per_leak_basis(consequences,
                                          num_leak_sizes,
                                          total_occupants):
    """
    Convert consequences to a leak basis, summing over all positions

    Parameters
    ----------
    num_leak_sizes : int
        Number of leak sizes

    total_occupants : int
        Number of occupants under consideration

    Returns
    -------
    consequence_probs_per_leak : array
        Probability of consequence per leak size
        (considering all occupants for each leak size)
    """
    # Convert from 1d to 2d where rows are consequences for single leak size
    consequence_probs_matrix = np.reshape(consequences, (num_leak_sizes, total_occupants))
    # Sum over positions so 1 val per leak size
    return np.sum(consequence_probs_matrix, axis=1)

def calculate_event_consequence(consequence_type,
                                num_leak_sizes,
                                total_occupants,
                                physical_responses,
                                consequence_modeling_decisions,
                                verbose=False):
    """
    Calculate consequence for different consequence types for all leak sizes and locations

    Parameters
    ----------
    consequence_type : str
        specifies what type of consequence model to use ('thermal', 'overp', or None)

    num_leak_sizes : int
        Number of leak sizes

    total_occupants : int
        Number of occupants under consideration

    physical_responses : dict
        Dictionary holding all physical responses associated with event

    consequence_modeling_decisions : dict
        Dictionary holding modeling decisions/specifications used to enacted the desired consequence model
    
    verbose : bool, optional
        Toogle extra output to command line

    Returns
    -------
    consequence_probs_per_leak : array
        Probability of specified consequence type per leak size
        (considering all occupants for each leak size)
    """
    if consequence_type == None:
        return np.array([0.0]*num_leak_sizes)
    elif consequence_type == 'thermal':
        consequence_model =  thermal_consequence
    elif consequence_type == 'overp':
        consequence_model = overpressure_consequence
    else:
        consequence_type_error_msg = ('Non valid consequence_type passed.'
                                    + ' Valid consequence_type inputs are'
                                    + ' "None", "thermal", and "overp".')
        raise ValueError(consequence_type_error_msg)

    if verbose:
        print(f"Calculating {consequence_type} fatality probabilities...")
    consequence_probs = consequence_model(physical_responses,
                                          consequence_modeling_decisions)
    consequence_probs_per_leak = convert_consequence_to_per_leak_basis(consequence_probs,
                                                                       num_leak_sizes,
                                                                       total_occupants)
    if verbose:
        print(f"Probit {consequence_type} data:\n{consequence_probs_per_leak}\n")
    return consequence_probs_per_leak


"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""
import unittest

import hyram.phys._utils as hpu


class TestGetDistanceToEffect(unittest.TestCase):
    """
    Test of function to get distance to a physical effect
    """
    @staticmethod
    def effect_function(x_vals, y_vals, z_vals, effect_val=1):
        """
        Dummy effect function that returns values for
        a 1/r decaying sphere centered at (1, 1, 0)
        with a dummy optional keyword argument that scales the values
        """
        origin = (1, 1, 0)
        effects = []
        for x_val, y_val, z_val in zip(x_vals, y_vals, z_vals):
            distance = ((x_val - origin[0]) ** 2
                        + (y_val - origin[1]) ** 2
                        + (z_val - origin[2]) ** 2) ** (1/2)
            if distance == 0:
                effect = 1e99
            else:
                effect = 1 / distance * effect_val
            effects.append(effect)
        return effects

    def test_effect_function(self):
        # check that both (1, 0, 0) and (0, 1, 0) are equal to 1
        self.assertEqual(self.effect_function([1], [0], [0]),
                         self.effect_function([0], [1], [0]))
        # check that maximum occurs at (1, 1, 0)
        x_vals = [1, 1.1, 1,  1, 1, 1, 0.9]
        y_vals = [1, 1, 1.1,  1, 0.9, 1, 1]
        z_vals = [0, 0, 0, -0.1, 0, 0.1, 0]
        effects = self.effect_function(x_vals, y_vals, z_vals)
        for non_max_effect in effects[1:]:
            self.assertGreater(effects[0], non_max_effect)

    def test_get_distance_to_effect(self):
        effect_value = 1
        from_point = (0, 1, 0)
        direction = 'x'
        effect_func = self.effect_function
        distance = hpu.get_distance_to_effect(effect_value, from_point, direction, effect_func)
        self.assertAlmostEqual(distance, 2, places=4)

    def test_get_distance_to_effect_with_kwarg(self):
        effect_value = 0.5
        from_point = (0, 1, 0)
        direction = 'x'
        effect_func = self.effect_function
        effect_scaling = 0.5
        distance = hpu.get_distance_to_effect(effect_value, from_point, direction, effect_func, effect_val=effect_scaling)
        self.assertAlmostEqual(distance, 2, places=4)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os

import numpy as np
import scipy as sp

from ._fuel_props import FuelProperties
from ._plots import plot_sliced_contour
from ._utils import get_distance_to_effect
from . import _overpressure_data as opdata

class Generic_overpressure_method:
    """
    Generic unconfined overpressure class

    Contains attributes and methods common to other
    unconfined overpressure sub-classes
    """
    data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def __init__(self, jet_object, heat_of_combustion=None, flammability_limits=None, origin_at_orifice=False):
        self.jet_object = jet_object
        self.heat_of_combustion = heat_of_combustion
        self.set_ambient_pressure()
        self.set_fuel_properties()
        self.set_flammable_mass(flammability_limits)
        self.set_overpressure_origin(origin_at_orifice)

    def set_ambient_pressure(self) -> None:
        self.ambient_pressure = self.jet_object.ambient.P  # Pa

    def set_fuel_properties(self) -> None:
        fuel_props = FuelProperties(self.jet_object.fluid.species)
        self.species = fuel_props.species
        nC = fuel_props.nC  # number of carbons (nC)
        # Mole fraction of fuel to mole fraction of air at stoichiometric condition
        self.fuel_to_air_stoich_ratio = 1/(4.76*(1+3*nC)/2)
        self.set_flammability_limits(fuel_props)
        self.set_heat_of_combustion(fuel_props)

    def set_flammability_limits(self, fuel_props) -> None:
        self.molar_lower_flammability_limit = fuel_props.LFL
        self.molar_upper_flammability_limit = fuel_props.UFL

    def set_flammable_mass(self, flammability_limits=None) -> None:
        if flammability_limits is not None:
            lower_flammability_limit = flammability_limits[0]
            upper_flammability_limit = flammability_limits[1]
        else:
            lower_flammability_limit = self.molar_lower_flammability_limit
            upper_flammability_limit = self.molar_upper_flammability_limit
        self.flammable_mass = self.jet_object.m_flammable(lower_flammability_limit,
                                                          upper_flammability_limit) # kg

    def set_heat_of_combustion(self, fuel_props) -> None:
        if self.heat_of_combustion == None:
            self.heat_of_combustion = fuel_props.dHc  # J/kg

    def set_overpressure_origin(self, origin_at_orifice:bool) -> None:
        if origin_at_orifice:
            self.origin = (0, 0, 0)
        else:
            # Overpressure-origin is the point at which the concentration
            # is mid-way between the lower and upper flamability limits
            LFL = self.molar_lower_flammability_limit
            UFL = self.molar_upper_flammability_limit
            mid_flammability = LFL + (UFL - LFL) / 2

            # Get jet streamline coordinate based on centerline concentration
            s_coord = np.interp(mid_flammability,
                                self.jet_object.X_cl[::-1],
                                self.jet_object.S[::-1])

            # Get x and y coordinates from jet based on streamline coordinate
            jet_x = np.interp(s_coord, self.jet_object.S, self.jet_object.x)
            jet_y = np.interp(s_coord, self.jet_object.S, self.jet_object.y)

            self.origin = (jet_x, jet_y, 0)

    @staticmethod
    def calc_distance(locations:list, origin) -> list:
        '''
        locations : list of locations
            List of locations of interest,
            each location is a tuple of 3 coordinates (in meters):
            [(x1, y1, z1), (x2, y2, z2), ...]
        origin : tuple or list
            origin location (x, y, z) of the overpressure event, in meters
        '''
        locations = np.array(locations)
        origin = np.array(origin)
        if len(locations) > 0:
            distance = np.linalg.norm(locations - origin, axis=1)
        else:
            distance = []
        return distance

    def calc_overpressure(self, locations:list) -> list:
        """
        Calculate overpressure

        Parameters
        ----------
        locations : list of locations
            List of locations at which to determine overpressure,
            each location is a tuple of 3 coordinates (in meters):
            [(x1, y1, z1), (x2, y2, z2), ...]

        Returns
        -------
        overpressure, float
            Overpressure in Pa
        """
        distance = self.calc_distance(locations=locations, origin=self.origin)
        scaled_distance = self.calc_scaled_distance(distance=distance)
        scaled_overpressure = self.get_scaled_overpressure(scaled_distance=scaled_distance)
        overpressure = self.calc_unscaled_overpressure(scaled_overpressure=scaled_overpressure)  # Pa
        return overpressure

    def calc_scaled_distance(self, distance):
        # Placeholder method; this will be over-written by each sub-class below
        scaled_distance = np.empty(len(distance))
        scaled_distance.fill(np.nan)
        return scaled_distance

    def get_scaled_overpressure(self, scaled_distance):
        # Placeholder method; this will be over-written by each sub-class below
        scaled_overpressure = np.empty(len(scaled_distance))
        scaled_overpressure.fill(np.nan)
        return scaled_overpressure

    def calc_unscaled_overpressure(self, scaled_overpressure):
        unscaled_overpressure = scaled_overpressure * self.ambient_pressure  # Pa
        return unscaled_overpressure

    def calc_impulse(self, locations:list) -> list:
        """
        Calculate impulse

        Parameters
        ----------
        locations : list of locations
            List of locations at which to determine impulse,
            each location is a tuple of 3 coordinates (in meters):
            [(x1, y1, z1), (x2, y2, z2), ...]

        Returns
        -------
        impulse, float
            Impulse in Pa*s
        """
        distance = self.calc_distance(locations=locations, origin=self.origin)
        scaled_distance = self.calc_scaled_distance(distance=distance)
        scaled_impulse = self.get_scaled_impulse(scaled_distance=scaled_distance)
        impulse = self.calc_unscaled_impulse(scaled_impulse=scaled_impulse)
        return impulse

    def get_scaled_impulse(self, scaled_distance):
        # Placeholder method; this will be over-written by each sub-class below
        scaled_impulse = np.empty(len(scaled_distance))
        scaled_impulse.fill(np.nan)
        return scaled_impulse

    def calc_unscaled_impulse(self, scaled_impulse):
        # Placeholder method; this will be over-written by each sub-class below
        unscaled_impulse = np.empty(len(scaled_impulse))
        unscaled_impulse.fill(np.nan)
        return unscaled_impulse

    def calculate_overpressure_for_list_of_locations(self, x:np.array, y:np.array, z:np.array) -> np.array:
        evaluation_locations = []
        x_y_shape = x.shape
        for x_value, y_value, z_location in zip(x.reshape(-1), y.reshape(-1), z.reshape(-1)):
            evaluation_locations += [np.array([x_value, y_value, z_location])]
        overpressures = self.calc_overpressure(evaluation_locations)
        overpressures = overpressures.reshape(x_y_shape)
        return overpressures

    def calculate_impulse_for_list_of_locations(self, x:np.array, y:np.array, z:np.array) -> np.array:
        evaluation_locations = []
        x_y_shape = x.shape
        for x_value, y_value, z_location in zip(x.reshape(-1), y.reshape(-1), z.reshape(-1)):
            evaluation_locations += [np.array([x_value, y_value, z_location])]
        impulses = self.calc_impulse(evaluation_locations)
        impulses = impulses.reshape(x_y_shape)
        return impulses

    def calc_distance_to_overpressure(self, overpressure, direction='x', negative_direction=False):
        """
        Calculate distance from leak point to a given overpressure in the direction specified

        Parameters
        ----------
        overpressure : float
            Overpressure in Pa
        direction : 'x', 'y', or 'z' (optional)
            Direction for which to calculate the distance
            (default is 'x')
        negative_direction : Boolean (optional)
            Whether or not to look in the negative direction instead of positive
            (default is False)

        Returns
        -------
        distance : float
            Real distance (m) to overpressure from leak-point
        """
        scaled_overpressure = self.calc_scaled_overpressure(overpressure=overpressure)  # unitless
        scaled_distance = self.get_scaled_distance_from_scaled_overpressure(scaled_overpressure=scaled_overpressure)
        distance_from_overpressure_origin = self.calc_unscaled_distance(scaled_distance=scaled_distance)  # m
        if direction == 'x':
            distance_from_leakpoint_to_overpressure_origin = self.origin[0]
        elif direction == 'y':
            distance_from_leakpoint_to_overpressure_origin = self.origin[1]
        elif direction == 'z':
            distance_from_leakpoint_to_overpressure_origin = self.origin[2]
        else:
            raise ValueError(f"Direction ('{direction}') must be 'x', 'y', or 'z'")
        if negative_direction:
            distance = distance_from_leakpoint_to_overpressure_origin - distance_from_overpressure_origin
        else:
            distance = distance_from_leakpoint_to_overpressure_origin + distance_from_overpressure_origin
        return distance

    def calc_scaled_overpressure(self, overpressure):
        scaled_overpressure = overpressure / self.ambient_pressure  # unitless
        return scaled_overpressure

    def get_scaled_distance_from_scaled_overpressure(self, scaled_overpressure, max_distance=500):
        # May be re-written by sub-classes below to use single interpolation rather than solver; 
        # currently calculates distance and scales it in order to match syntax of re-written methods below
        overpressure = self.calc_unscaled_overpressure(scaled_overpressure)
        distance = get_distance_to_effect(value=overpressure,
                                          from_point=self.origin,
                                          direction='x',  # does not matter, since overpressure is omni-directional
                                          effect_func=self.calculate_overpressure_for_list_of_locations,
                                          max_distance=max_distance)
        scaled_distance = self.calc_scaled_distance(distance)
        return scaled_distance

    def calc_unscaled_distance(self, scaled_distance):
        # Placeholder method; this may be over-written by each sub-class below
        distance = scaled_distance
        return distance

    def calc_distance_to_impulse(self, impulse, direction='x', negative_direction=False):
        """
        Calculate distance from leak point to a given impulse in the direction specified

        Parameters
        ----------
        impulse : float
            Impulse in Pa*s
        direction : 'x', 'y', or 'z', optional
            Direction for which to calculate the distance
            (default is 'x')
        negative_direction : Boolean (optional)
            Whether or not to look in the negative direction instead of positive
            (default is False)

        Returns
        -------
        distance : float
            Real distance (m) to impulse from leak-point
        """
        scaled_impulse = self.calc_scaled_impulse(impulse=impulse)
        scaled_distance = self.get_scaled_distance_from_scaled_impulse(scaled_impulse=scaled_impulse)
        distance_from_overpressure_origin = self.calc_unscaled_distance(scaled_distance=scaled_distance)  # m
        if direction == 'x':
            distance_from_leakpoint_to_overpressure_origin = self.origin[0]
        elif direction == 'y':
            distance_from_leakpoint_to_overpressure_origin = self.origin[1]
        elif direction == 'z':
            distance_from_leakpoint_to_overpressure_origin = self.origin[2]
        else:
            raise ValueError(f"Direction ('{direction}') must be 'x', 'y', or 'z'")
        if negative_direction:
            distance = distance_from_leakpoint_to_overpressure_origin - distance_from_overpressure_origin
        else:
            distance = distance_from_leakpoint_to_overpressure_origin + distance_from_overpressure_origin
        return distance

    def calc_scaled_impulse(self, impulse):
        # Placeholder method; this will be over-written by sub-classes below
        scaled_impulse = np.full_like(impulse, np.nan)
        return scaled_impulse

    def get_scaled_distance_from_scaled_impulse(self, scaled_impulse):
        # Placeholder method; this will be over-written by sub-classes below
        scaled_distance = np.full_like(scaled_impulse, np.nan)
        return scaled_distance

    def plot_overpressure_sliced(self, title=None,
                                 plot_filename='overpressure_sliced_contour.png',
                                 directory=os.getcwd(),
                                 contours=None,
                                 nx=50, ny=50, nz=50,
                                 xlims=None, ylims=None, zlims=None,
                                 savefig=False):
        '''
        Plots contour slices of overpressure levels

        Parameters
        ----------
        title: string (optional)
            title shown on plot
        plot_filename: string, optional
            file name to write
        directory: string, optional
            directory in which to save file
        contours: ndarray or list (optional)
            contour levels shown on plot (default values are from ICHS paper, safe setback distances)
            default levels correspond to glass breakage (5 kPa), people knocked over with some structural damage (16 kPa),
            and then significant fatalities and structural damage (70 kPa)
        nx, ny, nz: float (optional)
            number of points to solve for the overpressure in the x, y, and z directions
        xlims, ylims, zlims: tuples (optional)
            limits for x, y, and z axes
        savefig: Boolean (optional)
            determines if figure file is saved

        Returns
        -------
        If savefig is True, returns filename corresponding plot.
        If savefig is false, returns fig object.
        '''
        if contours is None:
            contours = [5, 16, 70]  # kPa

        colorbar_label = 'Overpressure [kPa]'

        fig_or_filepath = plot_sliced_contour(contours, xlims, ylims, zlims, nx, ny, nz,
                                              self.calc_distance_to_overpressure,
                                              self.calculate_overpressure_for_list_of_locations,
                                              self.origin, colorbar_label,
                                              title=title, savefig=savefig,
                                              directory=directory, filename=plot_filename)
        return fig_or_filepath

    def plot_impulse_sliced(self, title=None,
                            plot_filename='impulse_sliced_contour.png',
                            directory=os.getcwd(),
                            contours=None,
                            nx=50, ny=50, nz=50,
                            xlims=None, ylims=None, zlims=None,
                            savefig=False):
        '''
        Plots contour slices of impulse levels

        Parameters
        ----------
        title: string (optional)
            title shown on plot
        plot_filename: string, optional
            file name to write
        directory: string, optional
            directory in which to save file
        contours: ndarray or list (optional)
            contour levels to be shown on plot
            (default values are mean to be examples only;
            they are obtained by calculating the impulse value
            that would yield a 1% probability of fatality
            using the TNO Head Impact probit
            for the default peak overpressure values used:
            70, 16, and 5 kPa)
        nx, ny, nz: float (optional)
            number of points to solve for the impulse in the x, y, and z directions
        xlims, ylims, zlims: tuples (optional)
            limits for x, y, and z axes
        savefig: Boolean (optional)
            determines if figure file is saved

        Returns
        -------
        If savefig is True, returns filename corresponding plot.
        If savefig is false, returns fig object.
        '''
        if contours is None:
            contours = [0.13, 0.18, 0.27]  # kPa*s

        colorbar_label = 'Impulse [kPa*s]'

        fig_or_filepath = plot_sliced_contour(contours, xlims, ylims, zlims, nx, ny, nz,
                                              self.calc_distance_to_impulse,
                                              self.calculate_impulse_for_list_of_locations,
                                              self.origin, colorbar_label,
                                              title=title, savefig=savefig,
                                              directory=directory, filename=plot_filename)
        return fig_or_filepath


class BST_method(Generic_overpressure_method):
    """
    Vapor Cloud Explosions using BST Method
    based on flammable mass of fuel

    Parameters
    ----------
    jet_object : hyram Jet object
        Jet object representing fuel leak
    mach_flame_speed : float
        Mach flame speed: 0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2
        Detonation should use 5.2
    heat_of_combustion : float or None
        Heat of combustion of fuel (J/kg)
        Default is None: use default value for selected fuel
    flammability_limits : tuple or None
        Tuple of (lower_flammability_limit, upper_flammability_limit)
        Default is None: use default LFL, UFL for selected fuel
    origin_at_orifice : Boolean
        If True, the origin of the overpressure will be at the orifice
        Default is False: will calculate location of origin,
        the point at which the concentration in the unignited jet
        is mid-way between the default lower and upper flammability limits

    From: CCPS, "Guidelines for Vapor Cloud Explosion, Pressure Vessel Burst, BLEVE, and Flash Fire Hazards",
        Second Edition, Center for Chemical Process Safety, American Institute of Chemical Engineers,
        John Wiley & Sons, Inc.,2010
    """
    speed_of_sound = 340  # m/s

    def __init__(self, jet_object, mach_flame_speed,
                 heat_of_combustion=None, flammability_limits=None, origin_at_orifice=False):
        Generic_overpressure_method.__init__(self, jet_object, heat_of_combustion,
                                             flammability_limits, origin_at_orifice)
        self.set_mach_flame_speed(mach_flame_speed)
        self.scaled_peak_overpressure_data = opdata.scaled_peak_overpressure_data
        self.all_scaled_impulse_data = opdata.all_scaled_impulse_data
        self.energy = self.calc_energy()

    def set_mach_flame_speed(self, mach_flame_speed):
        """
        Change the mach flame speed

        Parameters
        ----------
        mach_flame_speed, float
            Mach flame speed: 0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2
            Detonation should use 5.2
        """
        self.mach_flame_speed = mach_flame_speed

    def calc_energy(self):
        ground_reflection_factor = 2
        energy = ground_reflection_factor * self.flammable_mass * self.heat_of_combustion  # J
        return energy

    def calc_scaled_distance(self, distance):
        scaled_distance = distance / (self.energy / self.ambient_pressure) ** (1/3)  # m/(J/Pa)^(1/3)
        return scaled_distance

    def get_scaled_overpressure(self, scaled_distance):
        scaled_distance_data = self.scaled_peak_overpressure_data['scaled_distance_Mf' + str(self.mach_flame_speed)]
        scaled_overpressure_data = self.scaled_peak_overpressure_data['scaled_overpressure_Mf' + str(self.mach_flame_speed)]
        scaled_overpressure = np.interp(x=scaled_distance,  # m/(J/Pa)^(1/3)
                                        xp=scaled_distance_data,  # m/(J/Pa)^(1/3)
                                        fp=scaled_overpressure_data)  # unitless
        return scaled_overpressure

    def get_scaled_impulse(self, scaled_distance):
        scaled_distance_data = self.all_scaled_impulse_data['scaled_distance_Mf' + str(self.mach_flame_speed)]
        scaled_impulse_data = self.all_scaled_impulse_data['scaled_impulse_Mf' + str(self.mach_flame_speed)]
        scaled_impulse = np.interp(x=scaled_distance,  # m/(J/Pa)^(1/3)
                                   xp=scaled_distance_data,  # m/(J/Pa)^(1/3)
                                   fp=scaled_impulse_data)  # Pa*s*(m/s)/J^(1/3)/Pa^(1/3)
        return scaled_impulse

    def calc_unscaled_impulse(self, scaled_impulse):
        unscaled_impulse = scaled_impulse * self.energy ** (1/3) * self.ambient_pressure ** (2/3) / self.speed_of_sound # Pa*s
        return unscaled_impulse

    def get_scaled_distance_from_scaled_overpressure(self, scaled_overpressure):
        scaled_distance_data = self.scaled_peak_overpressure_data['scaled_distance_Mf' + str(self.mach_flame_speed)]
        scaled_overpressure_data = self.scaled_peak_overpressure_data['scaled_overpressure_Mf' + str(self.mach_flame_speed)]
        scaled_distance = np.interp(x=scaled_overpressure,
                                    xp=scaled_overpressure_data[::-1],
                                    fp=scaled_distance_data[::-1])
        return scaled_distance

    def calc_unscaled_distance(self, scaled_distance):
        unscaled_distance = scaled_distance * (self.energy / self.ambient_pressure) ** (1/3)  # m
        distance = unscaled_distance + self.origin[0]  # account for x-location of origin
        return distance

    def calc_scaled_impulse(self, impulse):
        scaled_impulse = impulse * self.speed_of_sound / self.energy ** (1/3) / self.ambient_pressure ** (2/3)
        return scaled_impulse

    def get_scaled_distance_from_scaled_impulse(self, scaled_impulse):
        scaled_distance_data = self.all_scaled_impulse_data['scaled_distance_Mf' + str(self.mach_flame_speed)]
        scaled_impulse_data = self.all_scaled_impulse_data['scaled_impulse_Mf' + str(self.mach_flame_speed)]
        scaled_distance = np.interp(x=scaled_impulse,
                                    xp=scaled_impulse_data[::-1],
                                    fp=scaled_distance_data[::-1])
        return scaled_distance


class TNT_method(Generic_overpressure_method):
    """
    Vapor Cloud Explosions using TNT Equivalency Method
    based on flammable mass of fuel

    Assumes TNT blast energy of 4.68 MJ/kg

    Parameters
    ----------
    jet_object : HyRAM Jet object
        Jet object representing fuel leak
    equivalence_factor : float
        TNT mass equivalency factor, unitless
        CCPS source recommends a default value of 0.03
    heat_of_combustion : float or None
        Heat of combustion of fuel (J/kg)
        Default is None: use default value for selected fuel
    flammability_limits : tuple or None
        Tuple of (lower_flammability_limit, upper_flammability_limit)
        Default is None: use default LFL, UFL for selected fuel
    origin_at_orifice : Boolean
        If True, the origin of the overpressure will be at the orifice
        Default is False: will calculate location of origin,
        the point at which the concentration in the unignited jet
        is mid-way between the default lower and upper flammability limits

    From: CCPS, "Guidelines for Vapor Cloud Explosion, Pressure Vessel Burst, BLEVE, and Flash Fire Hazards",
        Second Edition, Center for Chemical Process Safety, American Institute of Chemical Engineers,
        John Wiley & Sons, Inc.,2010
    """
    def __init__(self, jet_object, equivalence_factor,
                 heat_of_combustion=None, flammability_limits=None, origin_at_orifice=False):
        Generic_overpressure_method.__init__(self, jet_object, heat_of_combustion,
                                             flammability_limits, origin_at_orifice)
        self.equivalence_factor = equivalence_factor  # unitless
        self.equiv_TNT_mass = self.calc_TNT_equiv_mass(self.equivalence_factor,
                                                       self.flammable_mass,
                                                       self.heat_of_combustion)  # kg
        self.scaled_peak_overP_data = opdata.scaled_peak_overP_data
        self.scaled_impulse_data = opdata.scaled_impulse_data

    @staticmethod
    def calc_TNT_equiv_mass(equivalence_factor, flammable_mass, heat_of_combustion):
        blast_energy_TNT = 4.68e6  # J/kg
        TNT_equiv_mass = equivalence_factor * flammable_mass * heat_of_combustion / blast_energy_TNT  # kg
        return TNT_equiv_mass

    def calc_scaled_distance(self, distance):
        scaled_distance = distance / self.equiv_TNT_mass ** (1/3)  # m/kg^(1/3)
        return scaled_distance

    def get_scaled_overpressure(self, scaled_distance):
        scaled_peak_overP_data = self.scaled_peak_overP_data
        scaled_overpressure = np.interp(x=scaled_distance,  # m/kg^(1/3)
                                        xp=scaled_peak_overP_data['scaled_distance'],  # m/kg^(1/3)
                                        fp=scaled_peak_overP_data['scaled_overpressure'])  # unitless
        return scaled_overpressure

    def get_scaled_impulse(self, scaled_distance):
        scaled_impulse_data = self.scaled_impulse_data
        scaled_impulse = np.interp(x=scaled_distance,  # m/kg^(1/3)
                                   xp=scaled_impulse_data['scaled_distance'],  # m/kg^(1/3)
                                   fp=scaled_impulse_data['scaled_impulse'])  # Pa*s/kg^(1/3)
        return scaled_impulse

    def calc_unscaled_impulse(self, scaled_impulse):
        unscaled_impulse = scaled_impulse * self.equiv_TNT_mass ** (1/3)  # Pa*s
        return unscaled_impulse

    def get_scaled_distance_from_scaled_overpressure(self, scaled_overpressure):
        scaled_peak_overP_data = self.scaled_peak_overP_data
        scaled_distance = np.interp(x=scaled_overpressure,  # unitless
                                    xp=scaled_peak_overP_data['scaled_overpressure'][::-1],
                                    fp=scaled_peak_overP_data['scaled_distance'][::-1])
        return scaled_distance

    def calc_unscaled_distance(self, scaled_distance):
        unscaled_distance = scaled_distance * self.equiv_TNT_mass ** (1/3)  # m
        distance = unscaled_distance + self.origin[0]  # account for x-location of origin
        return distance

    def calc_scaled_impulse(self, impulse):
        scaled_impulse = impulse / self.equiv_TNT_mass ** (1/3)  # Pa*s
        return scaled_impulse

    def get_scaled_distance_from_scaled_impulse(self, scaled_impulse):
        scaled_impulse_data = self.scaled_impulse_data
        scaled_distance = np.interp(x=scaled_impulse,
                                    xp=scaled_impulse_data['scaled_impulse'][::-1],
                                    fp=scaled_impulse_data['scaled_distance'][::-1])
        return scaled_distance


class Bauwens_method(Generic_overpressure_method):
    """
    Performs overpressure calculations - based on work of
    Bauwens and Dorofeev, ICHS 2019 paper ID 279:
    https://hysafe.info/uploads/2019_papers/279.pdf
    Bauwens and Dorofeev, CNF 2020:
    https://doi.org/10.1016/j.combustflame.2020.08.003

    Parameters
    ----------
    jet_object : hyram Jet object
        Jet object representing fuel leak
    heat_of_combustion : float or None
        Heat of combustion of fuel (J/kg)
        Default is None: use default value for selected fuel
    min_streamline_divisions : int
        Minimum number of divisions along streamline
         - increase for increased accuracy -
         normally just uses calculated S values
         (default=50)
    number_radial_divisions : int
        Number of divisions along radius
         - increase for increased accuracy
         (default=50)
    max_cell_gradient : float
        Maximum gradient in cell size across which
        detonations can propagate
        (default=0.1)
    minimum_number_detonable_cell : float
        Minimum number of cells for detonation propagation
        (default=5)
    origin_at_orifice : Boolean
        If True, the origin of the overpressure will be at the orifice
        Default is False: will calculate location of origin,
        the point at which the concentration in the unignited jet
        is mid-way between the default lower and upper flammability limits
    """
    def __init__(self, jet_object, heat_of_combustion=None, min_streamline_divisions=50, number_radial_divisions=50,
                 max_cell_gradient=0.1, minimum_number_detonable_cell=5, origin_at_orifice=False):
        self.jet_object = jet_object
        self.heat_of_combustion = heat_of_combustion  # J/kg
        self.set_ambient_pressure()  # Pa
        self.set_fuel_properties()
        self.set_overpressure_origin(origin_at_orifice)

        streamline_points, streamline_point_indices, streamline_point_interpolated_indices = self.calc_streamline_discretization(min_streamline_divisions)
        radial_coordinate_values, streamline_indice_values = self.calc_radial_and_streamline_meshgrid(number_radial_divisions, streamline_point_interpolated_indices)
        x_coordinate_values, y_coordinate_values = self.calc_spatial_discretization(radial_coordinate_values, streamline_indice_values, streamline_point_indices)
        moleFractionField, massFractionField, densityField = self.get_plume_mixture_properties(streamline_point_indices, streamline_indice_values, radial_coordinate_values)
        detonable_cell_size = self.calc_detonable_cell_size(moleFractionField)
        gradient_cell_size = self.calc_cell_size_gradient(x_coordinate_values, y_coordinate_values, detonable_cell_size)
        number_detonable_cells = self.calc_number_detonable_cells(moleFractionField, radial_coordinate_values, gradient_cell_size, detonable_cell_size, max_cell_gradient)
        self.detonable_mass = self.calc_detonable_mass(massFractionField, number_detonable_cells, minimum_number_detonable_cell, densityField, radial_coordinate_values, streamline_points)  # UNITS ? kg ?
        self.energy = self.calc_energy()

    def calc_streamline_discretization(self, min_streamline_divisions):
        if min_streamline_divisions > len(np.unique(self.jet_object.S)):
            streamline_points = np.logspace(np.log10(self.jet_object.S.min()), np.log10(self.jet_object.S.max()), min_streamline_divisions)
        else:
            streamline_points = np.unique(self.jet_object.S)
        streamline_point_indices = np.arange(len(self.jet_object.S))
        streamline_point_interpolated_indices = np.interp(streamline_points, self.jet_object.S, streamline_point_indices)
        return streamline_points, streamline_point_indices, streamline_point_interpolated_indices

    def calc_spatial_discretization(self, radial_coordinate_values, streamline_indice_values, streamline_point_indices):
        x_coordinate_values = np.interp(streamline_indice_values, streamline_point_indices, self.jet_object.x) + \
            radial_coordinate_values*np.sin(np.interp(streamline_indice_values, streamline_point_indices, self.jet_object.theta))
        y_coordinate_values = np.interp(streamline_indice_values, streamline_point_indices, self.jet_object.y) - \
            radial_coordinate_values*np.cos(np.interp(streamline_indice_values, streamline_point_indices, self.jet_object.theta))
        return x_coordinate_values, y_coordinate_values

    def calc_radial_and_streamline_meshgrid(self, number_radial_divisions, streamline_point_interpolated_indices):
        # Calculates logspaced points around 0 out to np.log10(3*np.max(self.B))
        # poshalf[::-1] just notation for reversing a numpy array
        poshalf = np.logspace(-5, np.log10(3*np.max(self.jet_object.B)), number_radial_divisions)
        radial_values = np.concatenate((-1 * poshalf[::-1], [0], poshalf))
        radial_coordinate_values, streamline_indice_values = np.meshgrid(radial_values, streamline_point_interpolated_indices)
        return radial_coordinate_values, streamline_indice_values

    def get_plume_mixture_properties(self, streamline_point_indices, streamline_coordinate_values, radial_coordinate_values):
        plumeHalfWidth = self.get_plume_halfwidth(streamline_coordinate_values, streamline_point_indices)
        centerlineDensity = self.get_centerline_density(streamline_coordinate_values, streamline_point_indices)
        centerlineMassFraction = self.get_centerline_massfraction(streamline_coordinate_values, streamline_point_indices)

        densityField = self.jet_object.ambient.rho + (centerlineDensity - self.jet_object.ambient.rho)*np.exp(-radial_coordinate_values**2/self.jet_object.lam**2/plumeHalfWidth**2)
        massFractionField = centerlineMassFraction*centerlineDensity*np.exp(-(radial_coordinate_values**2)/((self.jet_object.lam*plumeHalfWidth)**2))/densityField
        molecularWeightField = self.jet_object.ambient.therm.MW*self.jet_object.fluid.therm.MW/(massFractionField*(self.jet_object.ambient.therm.MW - self.jet_object.fluid.therm.MW) + self.jet_object.fluid.therm.MW)
        moleFractionField = massFractionField*molecularWeightField/self.jet_object.fluid.therm.MW

        return moleFractionField, massFractionField, densityField

    def get_plume_halfwidth(self, streamline_coordinate_values, streamline_point_indices):
        return np.interp(streamline_coordinate_values, streamline_point_indices, self.jet_object.B)

    def get_centerline_density(self, streamline_coordinate_values, streamline_point_indices):
        return np.interp(streamline_coordinate_values, streamline_point_indices, self.jet_object.rho_cl)

    def get_centerline_massfraction(self, streamline_coordinate_values, streamline_point_indices):
        return np.interp(streamline_coordinate_values, streamline_point_indices, self.jet_object.Y_cl)

    def calc_number_detonable_cells(self, moleFractionField, radial_coordinate_values, grad_cell_size,
                                    detonable_cell_size, max_cell_gradient):
        number_detonable_cells_1D = sp.integrate.simpson(1/detonable_cell_size, radial_coordinate_values, axis = 1)
        number_detonable_cells = np.zeros_like(detonable_cell_size)
        number_detonable_cells[:] = number_detonable_cells_1D[:, None]
        # check if detonable cells have sufficient mole fraction for detonation and that gradient isn't too large
        criteria1 = moleFractionField <= self.molar_lower_flammability_limit
        criteria2 = moleFractionField >= self.molar_upper_flammability_limit
        criteria3 = grad_cell_size > max_cell_gradient
        number_detonable_cells[criteria1 | criteria2| criteria3] = 0
        return number_detonable_cells

    def calc_cell_size_gradient(self, x_coordinate_values, y_coordinate_values, detonable_cell_size):
        return np.linalg.norm(np.gradient(detonable_cell_size, x_coordinate_values.T[0], y_coordinate_values[0]), axis = 0)

    def calc_detonable_mass(self, massFractionField, number_detonable_cells, minimum_number_detonable_cell, densityField,
                            radial_coordinate_values, streamline_points):
        massFractionField_det = np.copy(massFractionField)
        # determine plume locations where sufficient fuel for detonation or negative (integration only needs radius)
        massFractionField_det[(number_detonable_cells <= minimum_number_detonable_cell) | (radial_coordinate_values < 0)] = 0
        detonableFuelField = densityField*massFractionField_det*2*np.pi*radial_coordinate_values
        detonable_mass = sp.integrate.simpson(sp.integrate.simpson(detonableFuelField, radial_coordinate_values), streamline_points)
        return detonable_mass

    def calc_scaled_distance(self, distance):
        """
        Dimensionless standoff distance

        Parameters
        ----------
        distance, float
            Distance in meters

        Returns
        -------
        dimensionless_distance, float
            dimensionless standoff distance
        """
        # check if detonable mass exists
        condition_list = self.create_dimensionless_distance_condition_list(distance)
        # evaluate piecewise function for dimensionless distance (if conditionlist is true use function, else use inf)
        return np.piecewise(distance, condition_list, [self.dimensionless_distance, np.inf])

    def create_dimensionless_distance_condition_list(self, distance):
        condition1 = self.detonable_mass > np.zeros_like(distance)
        condition2 = self.detonable_mass <= np.zeros_like(distance)
        return [condition1, condition2]

    def dimensionless_distance(self, distance):
        return distance*(self.ambient_pressure/self.energy)**(1/3)

    def calc_energy(self):
        return self.detonable_mass*self.heat_of_combustion

    @staticmethod
    def calc_scaled_overpressure_from_dist(scaled_distance):
        '''Function to calculate scaled overpressure for Bauwens model based on scaled distance'''
        scaled_overpressure = (0.34 / scaled_distance ** (4/3)
                               + 0.062 / scaled_distance ** 2
                               + 0.0033 / scaled_distance ** 3)
        return scaled_overpressure

    def get_scaled_overpressure(self, scaled_distance):
        '''
        Determine scaled overpressure for Bauwens model based on scaled distance

        Parameters
        ----------
        scaled_distance : float
            scaled distance (unitless)
            any distance shorter than 1E-2 is set to 1E-2

        Returns
        ----------
        scaled_overpressure : float
            scaled overpressure (unitless)
        '''
        minimum_scaled_distance = 1E-2
        scaled_overpressure = []
        for scaled_dist in scaled_distance:
            if scaled_dist < minimum_scaled_distance:
                scaled_overp = self.calc_scaled_overpressure_from_dist(minimum_scaled_distance)
            else:
                scaled_overp = self.calc_scaled_overpressure_from_dist(scaled_dist)
            scaled_overpressure.append(scaled_overp)
        return np.array(scaled_overpressure)

    def calc_detonable_cell_size(self, moleFractionField):
        equivalence_ratio = moleFractionField/(1 - moleFractionField)/self.fuel_to_air_stoich_ratio
        molar_flammability_limits = np.array([self.molar_lower_flammability_limit, self.molar_upper_flammability_limit])
        (equivalence_ratio_lean_limit,
         equivalence_ratio_rich_limit) = (molar_flammability_limits /
                                         (1-molar_flammability_limits)/self.fuel_to_air_stoich_ratio)
        # check if equiv_ratio is outside of ER region or within it
        condition_list = self.create_cell_size_condition_list(equivalence_ratio, equivalence_ratio_lean_limit, equivalence_ratio_rich_limit)
        # evaluate piecewise function for equiv_ratio where if it is outside of ER range it is set to 1e99 otherwise it
        # Note: if np.inf is used rather than 1e99 - leads to RuntimeWarning in gradient
        detonable_cell_size = np.piecewise(equivalence_ratio,
                                           condition_list,
                                           [1e99, lambda equiv_ratio: self.detonation_cell_size(equiv_ratio)])
        return detonable_cell_size

    @staticmethod
    def create_cell_size_condition_list(equivalence_ratio, equivalence_ratio_lean_limit,
                                        equivalence_ratio_rich_limit):
        mixture_too_lean = equivalence_ratio < equivalence_ratio_lean_limit
        mixture_too_rich = equivalence_ratio > equivalence_ratio_rich_limit
        nonflamable_mixture = mixture_too_lean | mixture_too_rich

        above_lower_flamable_limit = equivalence_ratio >= equivalence_ratio_lean_limit
        below_lower_flamable_limit = equivalence_ratio <= equivalence_ratio_rich_limit
        flamable_mixture = above_lower_flamable_limit & below_lower_flamable_limit
        return [nonflamable_mixture, flamable_mixture]

    def detonation_cell_size(self, equiv_ratio):
        """
        Estimate detonation cell size from the equivalence ratio

        These estimations are done using fitted curves,
        see the Technical Reference Manual for details

        Parameters
        ----------
        equiv_ratio
            Equivalence ratio (unitless)

        Returns
        -------
        cell_size
            Detonation cell size (m)
        """
        dcl_fitted_params = {
            'H2': [2.94771698, -0.16536739, 2.2608031, -1.18064551, 0.45823461],
            'CH4': [5.768321, 1.13938677, 113.36802963, 0, 0],
            'C3H8': [4.44856885, -0.73108257, 5.50526263, 0, 0],
        }
        a, b, c, d, e = dcl_fitted_params[self.species]
        log_equiv_ratio = np.log(equiv_ratio)
        log_cell_size = (a
                         + b * log_equiv_ratio
                         + c * log_equiv_ratio ** 2
                         + d * log_equiv_ratio ** 3
                         + e * log_equiv_ratio ** 4)
        cell_size_mm = np.exp(log_cell_size)
        cell_size = cell_size_mm * sp.constants.milli
        return cell_size


class JallaisOverpressureH2(BST_method):
    """
    Calculate peak overpressure for hydrogen release
    using flammability limits and flame speed
    suggested by Jallais et al.
    https://doi.org/10.1002/prs.11965

    Parameters
    ----------
    jet_object : hyram Jet object
        Jet object representing fuel leak
    origin_at_orifice : Boolean
        If True, the origin of the overpressure will be at the orifice
        Default is False: will calculate location of origin,
        the point at which the concentration in the unignited jet
        is mid-way between the default lower and upper flamability limits
    """
    def __init__(self, jet_object, origin_at_orifice=False):
        if jet_object.fluid.species.lower() not in ['h2', 'hydrogen']:
            raise ValueError('JallaisOverpressureH2 is for hydrogen only')
        heat_of_combustion = None  # will use default heat of combustion value
        flammability_limits = (0.1, 0.75)  # from Jallais et al.
        mass_flow_rate = jet_object.mass_flow_rate
        self.flame_speed = self.calc_flame_speed(mass_flow_rate)
        self.calculated_mach_flame_speed = self.calc_mach_flame_speed(self.flame_speed)
        self.mach_flame_speed_curve = self.get_mach_flame_speed_curve(self.calculated_mach_flame_speed)
        BST_method.__init__(self, jet_object, self.mach_flame_speed_curve,
                            heat_of_combustion, flammability_limits, origin_at_orifice)

    @staticmethod
    def calc_flame_speed(mass_flow_rate):
        # From Jallais et al., for ignition at 30%:
        if mass_flow_rate < 0.5:  # kg/s
            flame_speed = 100  # m/s
        elif mass_flow_rate > 0.5 and mass_flow_rate < 1:  # kg/s
            flame_speed = 140  # m/s
        elif mass_flow_rate > 1 and mass_flow_rate < 10:  # kg/s
            flame_speed = 240  # m/s
        else:
            raise ValueError(f'Invalid jet mass flow rate: {mass_flow_rate}')
        return flame_speed

    def calc_mach_flame_speed(self, flame_speed):
        mach_flame_speed = flame_speed / self.speed_of_sound
        return mach_flame_speed

    @staticmethod
    def get_mach_flame_speed_curve(mach_flame_speed):
        valid_mach_flame_speeds = [0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2]
        # Get Mach flame speed from valid list that is closest to calculated value
        mach_flame_speed_curve = min(valid_mach_flame_speeds, key=lambda x: abs(x - mach_flame_speed))
        return mach_flame_speed_curve



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import math


def calc_pipe_inner_diameter(pipe_outer_diameter, pipe_wall_thickness):
    """
    Calculate inner diameter of pipe

    Parameters
    ----------
    pipe_outer_diameter : float
        Outer diameter of pipe

    pipe_wall_thickness : float
        Wall thickness of pipe
    
    Returns
    -------
    pipe_inner_diameter : float
        Inner diameter of pipe
    """
    pipe_inner_diam = pipe_outer_diameter - 2 * pipe_wall_thickness
    return pipe_inner_diam


def calc_pipe_flow_area(pipe_inner_diameter):
    """
    Calculate flow area of pipe

    Parameters
    ----------
    pipe_inner_diameter : float
        Inner diameter of pipe
    
    Returns
    -------
    pipe_flow_area : float
        Inner cross-sectional area of pipe
    """
    pipe_flow_area = math.pi * (pipe_inner_diameter / 2) ** 2
    return pipe_flow_area


def calc_orifice_diameter(pipe_flow_area, leak_size_fraction):
    """
    Calculate diameter of circular orifice
    based on a fractional leak size of a pipe flow area

    Parameters
    ----------
    pipe_flow_area : float
        Inner cross-sectional area of pipe

    leak_size_fraction : float
        Leak size in terms of fraction of pipe flow area

    Returns
    -------
    leak_diameter : float
        Diameter of circular leak orifice
    """
    leak_area = pipe_flow_area * leak_size_fraction
    leak_diameter = math.sqrt(4 * leak_area / math.pi)
    return leak_diameter



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

from hyram.qra import pipe_size


class TestPipeSizeCalcs(unittest.TestCase):
    """
    Test basic calculation of pipe sizes
    """
    def test_calc_pipe_inner_diameter(self):
        pipe_outer_diameter = 0.17  # m
        pipe_wall_thickness = 0.01  # m
        pipe_inner_diameter = pipe_size.calc_pipe_inner_diameter(pipe_outer_diameter,
                                                                 pipe_wall_thickness)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(pipe_inner_diameter, 0.15)

    def test_calc_pipe_flow_area(self):
        pipe_inner_diameter = 0.1  # m
        pipe_flow_area = pipe_size.calc_pipe_flow_area(pipe_inner_diameter)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(pipe_flow_area, 0.007853982)

    def test_calc_orifice_diameter(self):
        pipe_flow_area = 0.08  # m^2
        leak_size_fraction = 0.1
        leak_diameter = pipe_size.calc_orifice_diameter(pipe_flow_area,
                                                        leak_size_fraction)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(leak_diameter, 0.100925301)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""
import unittest

import numpy as np

from hyram.qra import component_set, leak_size_results


class LeakResultTestCase(unittest.TestCase):
    """ Evaluate creation of Leak Results objects. """

    def setUp(self):
        self.leak_sizes = [0.01, 0.1, 1, 10, 100]
        self.rel_species = 'h2'
        self.rel_phase = None
        self.comp_sets = component_set.init_component_sets(species=self.rel_species, phase=self.rel_phase, leak_sizes=self.leak_sizes)

    def test_leaks_without_overrides(self):
        overrides = [None, None, None, None, None]
        leak_results = leak_size_results.init_leak_results(self.leak_sizes, overrides, self.comp_sets)
        for i, result in enumerate(leak_results):
            self.assertGreater(np.sum(list(result.f_component_leaks.values())), 0)
            self.assertFalse(result.use_override)

            for component in self.comp_sets:
                # ignore when component unused
                if component.num_components:
                    self.assertTrue(result.f_release > 0)

    def test_leaks_with_overrides(self):
        overrides = [0.1, 0.2, 0.3, 0.4, 0.5]
        leak_results = leak_size_results.init_leak_results(self.leak_sizes, overrides, self.comp_sets)

        for i, result in enumerate(leak_results):
            self.assertEqual(result.f_release, overrides[i])
            self.assertTrue(result.use_override)
            self.assertEqual(np.sum(list(result.f_component_leaks.values())), 0)

    def test_leak_freqs_dont_match(self):
        overrides = [None, None, None, None, None]
        leak_results = leak_size_results.init_leak_results(self.leak_sizes, overrides, self.comp_sets)
        leak_000d01_freq = np.sum(list(leak_results[0].f_component_leaks.values()))
        leak_000d10_freq = np.sum(list(leak_results[1].f_component_leaks.values()))
        leak_001d00_freq = np.sum(list(leak_results[2].f_component_leaks.values()))
        leak_010d00_freq = np.sum(list(leak_results[3].f_component_leaks.values()))
        leak_100d00_freq = np.sum(list(leak_results[4].f_component_leaks.values()))

        self.assertNotAlmostEqual(leak_000d01_freq, leak_000d10_freq)
        self.assertNotAlmostEqual(leak_000d10_freq, leak_001d00_freq)
        self.assertNotAlmostEqual(leak_001d00_freq, leak_010d00_freq)
        self.assertNotAlmostEqual(leak_100d00_freq, leak_010d00_freq)

    def test_leaks_with_unchoked_mass_flows(self):
        leak_results = leak_size_results.init_leak_results(self.leak_sizes, None, self.comp_sets, mass_flow_leak_size=1, mass_flow=0.44)
        flows = [0.44, 4.4, 44., 440, 4400]
        for i, result in enumerate(leak_results):
            self.assertAlmostEqual(result.mass_flow_override, flows[i], places=4)

        leak_results = leak_size_results.init_leak_results(self.leak_sizes, None, self.comp_sets, mass_flow_leak_size=10, mass_flow=5.9)
        flows = [0.59, 5.9, 59, 590, 5900]
        for i, result in enumerate(leak_results):
            self.assertAlmostEqual(result.mass_flow_override, flows[i], places=4)

        leak_results = leak_size_results.init_leak_results(self.leak_sizes, None, self.comp_sets, mass_flow_leak_size=100, mass_flow=0.44)
        flows = [0.0044, 0.044, 0.44, 4.4, 44.]
        for i, result in enumerate(leak_results):
            self.assertAlmostEqual(result.mass_flow_override, flows[i], places=4)

        leak_results = leak_size_results.init_leak_results(self.leak_sizes, None, self.comp_sets, mass_flow_leak_size=1000, mass_flow=0.01)
        flows = [0.00001, 0.0001, 0.001, 0.01, 0.1]
        for i, result in enumerate(leak_results):
            self.assertAlmostEqual(result.mass_flow_override, flows[i], places=4)

        leak_results = leak_size_results.init_leak_results(self.leak_sizes, None, self.comp_sets, mass_flow_leak_size=10000, mass_flow=1)
        flows = [0.0001, 0.001, 0.01, 0.1, 1]
        for i, result in enumerate(leak_results):
            self.assertAlmostEqual(result.mass_flow_override, flows[i], places=4)

    def test_leaks_with_unrecognized_mass_flow_size_raises_error(self):
        with self.assertRaises(ValueError):
            leak_results = leak_size_results.init_leak_results(self.leak_sizes, None, self.comp_sets, mass_flow_leak_size=4, mass_flow=4)

        with self.assertRaises(ValueError):
            leak_results = leak_size_results.init_leak_results(self.leak_sizes, None, self.comp_sets, mass_flow_leak_size=0.1, mass_flow=4)



"""
Copyright 2015-2023 National Techno`log`y & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import unittest
import json

import numpy as np
from scipy.constants import milli
from scipy.optimize import fsolve

from hyram.phys import Fluid, Orifice, Source, Vent, Enclosure, IndoorRelease
from hyram.utilities import misc_utils

import utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

# Flags to enable command line output, pyplots, and text file output
VERBOSE = False
CREATE_PLOTS = False
CREATE_OUTPUT = False

# Absolute paths to input/output data
DATA_LOC = os.path.join(os.path.dirname(__file__), 'data')
OUTPUT_LOC = os.path.join(misc_utils.get_temp_folder(), 'validation-overpressure_accumulation')
OUTPUT_FILE = os.path.join(OUTPUT_LOC, 'Validation.txt')
LIMITS_FILE = os.path.join(DATA_LOC, 'Limits-overpressure_accumulation.json')


class Test_Ekoto_2012(unittest.TestCase):
    """
    Test mole fraction and peak pressure calculations against Ekoto et al. (2012).
    """

    def setUp(self):
        self.amb_species = "air"
        self.rel_species = "H2"
        self.amb_temp = 297
        self.amb_pressure = 101325
        self.rel_pressure = 13450000
        self.rel_temp = 297
        self.tank_volume = .00363  # m3
        self.orifice_diameter = .00356  # m
        self.orifice_dis_coeff = 0.75
        self.rel_dis_coeff = 1
        self.rel_area = 0.0171  # m2
        self.rel_height = 0.2495  # m
        self.enclosure_height = 2.72  # m
        self.floor_ceil_area = 16.722  # m2
        self.dist_rel_to_wall = 2.1255  # m
        self.rel_angle = 0
        self.is_steady = False

        fluid = Fluid(species=self.rel_species, P=self.rel_pressure, T=self.rel_temp)
        self.ambient = Fluid(species=self.amb_species, T=self.amb_temp, P=self.amb_pressure)
        self.orifice = Orifice(self.orifice_diameter, Cd=self.orifice_dis_coeff)
        self.source = Source(self.tank_volume, fluid)

        # Create IndoorRelease object with no vent
        ceil_vent_xarea = 0.00865  # m2
        ceil_vent_height = 2.42
        floor_vent_xarea = 0.0
        floor_vent_height = 0.0
        vol_flow_rate = 0.0

        ceil_vent = Vent(ceil_vent_xarea, ceil_vent_height, self.rel_dis_coeff, vol_flow_rate)
        floor_vent = Vent(floor_vent_xarea, floor_vent_height, self.rel_dis_coeff, vol_flow_rate)
        enclosure = Enclosure(self.enclosure_height, self.floor_ceil_area, self.rel_height, ceil_vent, floor_vent,
                              self.dist_rel_to_wall)

        self.novent_release = IndoorRelease(self.source, self.orifice, self.ambient, enclosure, tmax=20.5,release_area=self.rel_area,
                                            theta0=self.rel_angle, verbose=False)

        # Create IndoorRelease object with vent
        ceil_vent_xarea = 0.1  # m2
        vol_flow_rate = 6.5 / 60  # m3/s

        ceil_vent = Vent(ceil_vent_xarea, ceil_vent_height, self.rel_dis_coeff, vol_flow_rate)
        floor_vent = Vent(floor_vent_xarea, floor_vent_height, self.rel_dis_coeff, vol_flow_rate)
        enclosure = Enclosure(self.enclosure_height, self.floor_ceil_area, self.rel_height, ceil_vent, floor_vent,
                                            self.dist_rel_to_wall)

        self.vent_release = IndoorRelease(self.source, self.orifice, self.ambient, enclosure, tmax=20.5,
                                    release_area=self.rel_area, theta0=self.rel_angle, verbose=False)


        # Import experimental data
        self.data = utils.read_csv(os.path.join(DATA_LOC, 'ekoto-2012-fig9.csv'), header=2)

        # Import error limits
        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    # Fig 9, Mole Fraction vs. Time
    def test_fig9_noVent(self):
        sensor_list = ['S04 No Vent', 'S08 No Vent', 'S11 No Vent', 'S04 No Vent 2', 'S08 No Vent 2', 'S11 No Vent 2']
        for i, sensor in enumerate(sensor_list):
            title = f'Ekoto et al. 2012 - Figure 9 Concentration vs. Time, {sensor}'
            error_limits = self.limits[title]

            # Get data from csv
            time = self.data[f'x{i+1}']
            exp_concentration = self.data[f'y{i+1}']

            # Calculate concentration
            calc_concentration = self.novent_release.concentration(time) / 100

            error = utils.get_error(exp_vals=exp_concentration,
                                    calc_vals=calc_concentration,
                                    error_limits=error_limits,
                                    units='Mole Fraction',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            # Create plots
            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                   x=time,
                                   exp_y=exp_concentration,
                                   calc_y=calc_concentration,
                                   max_error=error['Max Absolute Error'],
                                   avg_error=error['Avg Absolute Error'],
                                   smape=error['SMAPE'],
                                   title=title,
                                   xlabel='Time (s)',
                                   ylabel='Mole Fraction')

    # Fig 9, w/ vent
    def test_fig9_vent(self):
        sensor_list = ['S04 Vent', 'S08 Vent', 'S11 Vent']
        for i, sensor in enumerate(sensor_list):
            title = f'Ekoto et al. 2012 - Figure 9 Concentration vs. Time, {sensor}'
            error_limits = self.limits[title]

            # Get data from csv
            time = self.data[f'x{7+i}']
            exp_concentration = self.data[f'y{7+i}']

            # Calculate concentration
            calc_concentration = self.vent_release.concentration(time) / 100

            error = utils.get_error(exp_vals=exp_concentration,
                                    calc_vals=calc_concentration,
                                    error_limits=error_limits,
                                    units='Mole Fraction',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            # Create plots
            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                   x=time,
                                   exp_y=exp_concentration,
                                   calc_y=calc_concentration,
                                   max_error=error['Max Absolute Error'],
                                   avg_error=error['Avg Absolute Error'],
                                   smape=error['SMAPE'],
                                   title=title,
                                   xlabel='Time (s)',
                                   ylabel='Mole Fraction')

    # Fig 14, Test peak pressure. As point-to-point comparisons, this test uses its own error checking
    def test_peak_pressure(self):
        amb_species = "air"
        rel_species = "H2"
        amb_temp = 297
        amb_pressure = 101325
        rel_pressure = 13450000
        rel_temp = 297
        tank_volume = .00363  # m3
        orifice_diameter = .00356  # m
        orifice_dis_coeff = 0.75
        rel_dis_coeff = 1
        rel_area = 0.0171  # m2
        rel_height = 0.2495  # m
        enclosure_height = 2.72  # m
        floor_ceil_area = 16.722  # m2
        dist_to_wall = 2.1255  # m
        rel_angle = 0
        xloc = 0
        yloc = 0

        fluid = Fluid(species=rel_species, P=rel_pressure, T=rel_temp)
        ambient = Fluid(species=amb_species, T=amb_temp, P=amb_pressure)
        orifice = Orifice(orifice_diameter, Cd=orifice_dis_coeff)
        source = Source(tank_volume, fluid)

        # From the notebook
        t = 3

        # Lumping these 3 into this one test since they're all just point-to-point comparisons
        # Test 12 - Small vent, passive ventilation
        title = "Ekoto et al. 2012 - Figure 14, Peak Pressure Test 12, Small vent w/ passive ventilation"
        error_limits = self.limits[title]

        ceil_vent = Vent(0.00368, 2.42, rel_dis_coeff, 0)
        floor_vent = Vent(0, 0, rel_dis_coeff, 0)
        enclosure = Enclosure(enclosure_height, floor_ceil_area, rel_height, ceil_vent, floor_vent,
                              dist_to_wall)
        release = IndoorRelease(source, orifice, ambient, enclosure, tmax=8.5,
                                release_area=rel_area, theta0=rel_angle,
                                x0=xloc, y0=yloc, verbose=False)
	
        exp_pressure = 24.48451643
        calc_pressure = release.pressure(t) / 1000
        error = utils.get_error(exp_vals=[exp_pressure],
                                calc_vals=[calc_pressure],
                                error_limits=error_limits,
                                units='kPa',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        # Only test absolute error and percent error values
        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Percent Error out of range")

        # Test 11 - Big vent, passive ventilation
        title = "Ekoto et al. 2012 - Figure 14, Peak Pressure Test 11, Big vent w/ passive ventilation"
        error_limits = self.limits[title]

        ceil_vent = Vent(0.09716, 2.42, rel_dis_coeff, 0)
        floor_vent = Vent(0, 0, rel_dis_coeff, 0)
        enclosure = Enclosure(enclosure_height, floor_ceil_area, rel_height, ceil_vent, floor_vent,
                              dist_to_wall)
        release = IndoorRelease(source, orifice, ambient, enclosure, tmax=8.5,
                                release_area=rel_area, theta0=rel_angle,
                                x0=xloc, y0=yloc, verbose=False)

        exp_pressure = 4.131355932
        calc_pressure = release.pressure(t) / 1000
        error = utils.get_error(exp_vals=[exp_pressure],
                                calc_vals=[calc_pressure],
                                error_limits=error_limits,
                                units='kPa',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        # Only test absolute error and percent error values
        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Percent Error out of range")


        # Test 10 - Big vent, active ventilation
        title = "Ekoto et al. 2012 - Figure 14, Peak Pressure Test 10, Big vent w/ active ventilation"
        error_limits = self.limits[title]

        ceil_vent = Vent(0.09716, 2.42, rel_dis_coeff, 6.3/60)
        floor_vent = Vent(0, 0, rel_dis_coeff, 6.3/60)
        enclosure = Enclosure(enclosure_height, floor_ceil_area, rel_height, ceil_vent, floor_vent,
                                    dist_to_wall)
        release = IndoorRelease(source, orifice, ambient, enclosure, tmax=8.5,
                                release_area=rel_area, theta0=rel_angle,
                                x0=xloc, y0=yloc, verbose=False)

        exp_pressure = 3.019067797
        calc_pressure = release.pressure(t) / 1000
        error = utils.get_error(exp_vals=[exp_pressure],
                                calc_vals=[calc_pressure],
                                error_limits=error_limits,
                                units='kPa',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        # Only test absolute error and percent error values
        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Percent Error out of range")


class Test_Giannissi_2015(unittest.TestCase):
    """
    Test concentration calculation against Giannissi al. (2015)
    """

    def setUp(self):
        amb_species = "air"
        rel_species = "H2"
        amb_temp = 287
        amb_pressure = 101325
        rel_pressure = 1.70e06
        rel_temp = 285
        tank_volume = 3943 / 1000 / 60 # std liters/min to m3/s
        orifice_diameter = .00055  # m
        orifice_dis_coeff = 1
        rel_dis_coeff = 1
        rel_area = np.pi * (orifice_diameter / 2) ** 2  # m2
        rel_height = 0.5  # m
        enclosure_height = 2.5  # m
        floor_ceil_area = 12.5  # m2
        dist_to_wall = 2.5  # m
        rel_angle = np.pi / 2
        is_steady = True

        # Set max time to pressure peak
        max_time = 1600

        ceil_vent_xarea = 0.2241  # m2
        ceil_vent_height = 2.13
        floor_vent_xarea = 0.0
        floor_vent_height = 0.0
        vol_flow_rate = 0.0

        fluid = Fluid(species=rel_species, P=rel_pressure, T=rel_temp)
        ambient = Fluid(species=amb_species, T=amb_temp, P=amb_pressure)
        orifice = Orifice(orifice_diameter, Cd=orifice_dis_coeff)
        source = Source(tank_volume, fluid)
        ceil_vent = Vent(ceil_vent_xarea, ceil_vent_height, rel_dis_coeff, vol_flow_rate)
        floor_vent = Vent(floor_vent_xarea, floor_vent_height, rel_dis_coeff, vol_flow_rate)
        enclosure = Enclosure(enclosure_height, floor_ceil_area, rel_height, ceil_vent, floor_vent,
                                            dist_to_wall)
        self.release = IndoorRelease(source, orifice, ambient, enclosure, tmax=max_time,
                                        release_area=rel_area, theta0=rel_angle, steady=is_steady, nsteady = 20,
                                        x0=0, y0=rel_height, verbose=False)

        self.data = utils.read_csv(os.path.join(DATA_LOC, 'giannissi-2015-fig7.csv'), header=2)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    def test_concentration_time_s18(self):
        idx = 0
        title = "Giannissi et al. 2015 - Figure 7, Concentration v. Time, Sensor 18"
        error_limits = self.limits[title]

        # Cull data past 1600s
        time = self.data[f'x{idx+1}'][:16]
        exp_concentration = self.data[f'y{idx+1}'][:16]

        calc_concentration = self.release.concentration(time)

        error = utils.get_error(exp_vals=exp_concentration,
                                calc_vals=calc_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=time,
                                exp_y=exp_concentration,
                                calc_y=calc_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Time (s)',
                                ylabel='Concentration (vol %)')

    def test_concentration_time_s21(self):
        idx = 1
        title = "Giannissi et al. 2015 - Figure 7, Concentration v. Time, Sensor 21"
        error_limits = self.limits[title]

        # Cull data past 1600s
        time = self.data[f'x{idx+1}'][:16]
        exp_concentration = self.data[f'y{idx+1}'][:16]

        calc_concentration = self.release.concentration(time)

        error = utils.get_error(exp_vals=exp_concentration,
                                calc_vals=calc_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=time,
                                exp_y=exp_concentration,
                                calc_y=calc_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Time (s)',
                                ylabel='Concentration (vol %)')

    def test_concentration_time_s27(self):
        idx = 0
        title = "Giannissi et al. 2015 - Figure 7, Concentration v. Time, Sensor 27"
        error_limits = self.limits[title]

        # Cull data past 1600s
        time = self.data[f'x{idx+1}'][:16]
        exp_concentration = self.data[f'y{idx+1}'][:16]

        calc_concentration = self.release.concentration(time)

        error = utils.get_error(exp_vals=exp_concentration,
                                calc_vals=calc_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=time,
                                exp_y=exp_concentration,
                                calc_y=calc_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Time (s)',
                                ylabel='Concentration (vol %)')


class Test_Merilo(unittest.TestCase):
    """
    Test concentration calculation against Merilo al. (2011) Figure 2
    """

    def setUp(self):
        self.amb_species = "air"
        self.rel_species = "H2"
        self.amb_temp = 285
        self.amb_pressure = 101325
        self.rel_temp = 285
        self.tank_volume = 1e6 # unnaturally large so that flow is nominally steady
        release_diameter = 7.75*milli  # m
        self.orifice_dis_coeff = 1
        self.rel_dis_coeff = 1
        self.rel_area = np.pi * (release_diameter / 2) ** 2  # m2
        self.enclosure_height = 2.72
        self.floor_ceil_area = 22.204
        self.rel_angle = np.pi / 2
        self.is_steady = True # ESH
        self.floor_vent_flow = 0.0

        self.ceil_vent_xarea = 0.11  # m2
        self.ceil_vent_height = 2.42
        self.floor_vent_xarea = .11
        self.floor_vent_height = .17

        tests = [
            {'name': 'Test 1', 'rel_rate': 9.22, 'mass_rel': 3.07,  'vel': 668, 'dur': 20, 'rel_ht': 0.25, 'dist_to_wall': 3.04, 'flow': 0, 'd_throat':1.7*milli, 'tmax':20*60},
            {'name': 'Test 2', 'rel_rate': 9.04, 'mass_rel': 3.01,  'vel': 653, 'dur': 20, 'rel_ht': 1, 'dist_to_wall': 4.85, 'flow': 0, 'd_throat':1.7*milli, 'tmax':20*60},
            {'name': 'Test 3', 'rel_rate': 0.88, 'mass_rel': .44,  'vel': 63, 'dur': 30, 'rel_ht': 1, 'dist_to_wall': 4.85, 'flow': 0, 'd_throat':1.2*milli, 'tmax':30*60},
            {'name': 'Test 7', 'rel_rate': 6.7, 'mass_rel': 4.47,  'vel': 502, 'dur': 30, 'rel_ht': .25, 'dist_to_wall': 3.04, 'flow': 0.1, 'd_throat':1.7*milli, 'tmax':45*60},
        ]

        self.releases = []
        for test in tests:
            # Set up release
            ambient = Fluid(species=self.amb_species, T=self.amb_temp, P=self.amb_pressure)
            orifice = Orifice(test['d_throat'])

            def err(P):
                g = Fluid(species=self.rel_species, P=P, T=self.rel_temp)
                return (test['rel_rate'] / 3600) - orifice.mdot(orifice.flow(g, self.amb_pressure))
            pressure = fsolve(err, 10*self.amb_pressure)[0]

            fluid = Fluid(species=self.rel_species, P=pressure, T=self.rel_temp)
            orifice = Orifice(test['d_throat'], Cd=self.orifice_dis_coeff)
            source = Source(self.tank_volume, fluid)
            ceil_vent = Vent(self.ceil_vent_xarea, self.ceil_vent_height, self.rel_dis_coeff, test['flow'])
            floor_vent = Vent(self.floor_vent_xarea, self.floor_vent_height, self.rel_dis_coeff, self.floor_vent_flow)
            enclosure = Enclosure(self.enclosure_height, self.floor_ceil_area, test['rel_ht'], ceil_vent, floor_vent, test['dist_to_wall'])
            release = IndoorRelease(source, orifice, ambient, enclosure, tmax=test['tmax'],
                                    release_area=self.rel_area, theta0=self.rel_angle, steady=self.is_steady,
                                    x0=0, y0=enclosure.H_release, verbose=False, nsteady=30)
            self.releases.append(release)

        self.data = utils.read_csv(os.path.join(DATA_LOC, 'merilo-2011-fig2.csv'), header=2)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    def test_concentration_time_test1(self):
        idx = 0
        title = f"Merilo et al. 2011 - Figure 2, Concentration v. Time, Test 1"
        error_limits = self.limits[title]

        # Using best fit line from experimental data
        time = self.data[f'x{(idx+1)*4}']
        exp_concentration = self.data[f'y{(idx+1)*4}']

        # Use linear interpolation on the calculated values
        calc_concentration = np.interp(time, self.releases[idx].t_layer/60, self.releases[idx].x_layer*100)

        error = utils.get_error(exp_vals=exp_concentration,
                                calc_vals=calc_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=time,
                                exp_y=exp_concentration,
                                calc_y=calc_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Time (s)',
                                ylabel='Concentration (vol %)')

    def test_concentration_time_test2(self):
        idx = 1
        title = f"Merilo et al. 2011 - Figure 2, Concentration v. Time, Test 2"
        error_limits = self.limits[title]

        time = self.data[f'x{(idx+1)*4}']
        exp_concentration = self.data[f'y{(idx+1)*4}']
        calc_concentration = np.interp(time, self.releases[idx].t_layer/60, self.releases[idx].x_layer*100)

        error = utils.get_error(exp_vals=exp_concentration,
                                calc_vals=calc_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=time,
                                exp_y=exp_concentration,
                                calc_y=calc_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Time (s)',
                                ylabel='Concentration (vol %)')

    def test_concentration_time_test3(self):
        idx = 2
        title = f"Merilo et al. 2011 - Figure 2, Concentration v. Time, Test 3"
        error_limits = self.limits[title]

        time = self.data[f'x{(idx+1)*4}']
        exp_concentration = self.data[f'y{(idx+1)*4}']
        calc_concentration = np.interp(time, self.releases[idx].t_layer/60, self.releases[idx].x_layer*100)

        error = utils.get_error(exp_vals=exp_concentration,
                                calc_vals=calc_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=time,
                                exp_y=exp_concentration,
                                calc_y=calc_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Time (s)',
                                ylabel='Concentration (vol %)')

    def test_concentration_time_test7(self):
        idx = 3
        title = f"Merilo et al. 2011 - Figure 2, Concentration v. Time, Test 7"
        error_limits = self.limits[title]

        time = self.data[f'x{(idx+1)*4}']
        exp_concentration = self.data[f'y{(idx+1)*4}']
        calc_concentration = np.interp(time, self.releases[idx].t_layer/60, self.releases[idx].x_layer*100)

        error = utils.get_error(exp_vals=exp_concentration,
                                calc_vals=calc_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=time,
                                exp_y=exp_concentration,
                                calc_y=calc_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Time (s)',
                                ylabel='Concentration (vol %)')


class Test_BMHA_Fig7(unittest.TestCase):
    """
    Test concentration vs flowrate calculation against Bernard-Michel and Houssin-Agbomson (2017), Figure 7 -- release in a 2m^3 enclosure
    """

    def setUp(self):
        amb_species = "air"
        rel_species = "H2"
        amb_temp = 286
        amb_pressure = 101325
        rel_temp = 286
        tank_volume = 1e6 # unnaturally large so that flow is nominally steady
        rel_dis_coeff = 1

        diameter_1 = .0272
        diameter_2 = .004
        area_1 = np.pi * (diameter_1 / 2) ** 2  # m2
        area_2 = np.pi * (diameter_2 / 2) ** 2  # m2
        Qs2m3 = np.array([5.2, 21, 73, 218])/1000/60
        rel_angle = np.pi / 2
        x0 = 0
        y0 = 0

        # Figure 7
        rel_height = 0.27  # m
        enclosure_height = 2.1  # m
        floor_ceil_area = .9216
        dist_to_wall = 0.48  # m
        ceil_vent_xarea = 0.1862  # m2
        ceil_vent_height = 1.7
        floor_vent_xarea = .1862
        floor_vent_height = 0.02
        is_steady = True

        d_imaginary = 2e-4
        ambient = Fluid(species=amb_species, T=amb_temp, P=amb_pressure)
        orifice = Orifice(d_imaginary)
        rho_H2_STP = Fluid(species='hydrogen', T=295, P=101325).rho

        self.flowrate = Qs2m3 * 60 * 1000

        # Calculate pressure data
        pressures = []
        for Q in Qs2m3:
            def err_hy3(P):
                g = Fluid(species=rel_species, P=P, T=rel_temp)
                throat = orifice.flow(g, amb_pressure)
                return Q - orifice.mdot(throat)/rho_H2_STP

            pressures.append(fsolve(err_hy3, 10*amb_pressure)[0])

        data = utils.read_csv(os.path.join(DATA_LOC, 'bernardmichelhoussinagbomson-2017-fig7.csv'), header=2)

        # Create concentration data for both release areas, d=4mm and d=27mm
        self.calc_4mm_concentration = []
        self.calc_27mm_concentration = []
        for pressure in pressures:
            fluid = Fluid(species=rel_species, P=pressure, T=rel_temp)
            source = Source(tank_volume, fluid)

            ceil_vent = Vent(ceil_vent_xarea, ceil_vent_height, rel_dis_coeff, 0)
            floor_vent = Vent(floor_vent_xarea, floor_vent_height, rel_dis_coeff, 0)
            enclosure = Enclosure(enclosure_height, floor_ceil_area, rel_height, ceil_vent, floor_vent,
                                        dist_to_wall)

            release_4mm = IndoorRelease(source, orifice, ambient, enclosure,
                                        release_area=area_1, theta0=rel_angle, steady=is_steady,
                                        x0=x0, y0=y0, verbose=False)
            release_27mm = IndoorRelease(source, orifice, ambient, enclosure,
                                         release_area=area_2, theta0=rel_angle, steady=is_steady,
                                         x0=x0, y0=y0, verbose=False)

            self.calc_4mm_concentration.append(release_4mm.x_layer[-1] * 100)
            self.calc_27mm_concentration.append(release_27mm.x_layer[-1] * 100)

        self.exp_4mm_concentration = data['y1']
        self.exp_27mm_concentration = data['y2']

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    def test_flowrate_4mm(self):
        title = "Bernard et al. 2017 - Figure 7, Concentration v. Flow Rate, d = 4mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=self.exp_4mm_concentration,
                                calc_vals=self.calc_4mm_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=self.flowrate,
                                exp_y=self.exp_4mm_concentration,
                                calc_y=self.calc_4mm_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Flow Rate (NL/min)',
                                ylabel='Concentration (vol %)')

    def test_flowrate_27mm(self):
        title = "Bernard et al. 2017 - Figure 7, Concentration v. Flow Rate, d = 27mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=self.exp_27mm_concentration,
                                calc_vals=self.calc_27mm_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=self.flowrate,
                                exp_y=self.exp_27mm_concentration,
                                calc_y=self.calc_27mm_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Flow Rate (NL/min)',
                                ylabel='Concentration (vol %)')


class Test_BMHA_Fig9(unittest.TestCase):
    """
    Test concentration vs flowrate calculation against Bernard-Michel and Houssin-Agbomson (2017), Figure 9 -- release in a 1m^3 enclosure
    """

    def setUp(self):
        amb_species = "air"
        rel_species = "H2"
        amb_temp = 286
        amb_pressure = 101325
        rel_temp = 286
        tank_volume = 1e6 # unnaturally large so that flow is nominally steady
        rel_dis_coeff = 1

        diameter_1 = .0272
        diameter_2 = .004
        area_1 = np.pi * (diameter_1 / 2) ** 2  # m2
        area_2 = np.pi * (diameter_2 / 2) ** 2  # m2
        Qs1m3 = np.array([10, 21, 62, 104, 218])/1000/60
        rel_height = .08
        enclosure_height = 1
        floor_ceil_area = .990025
        dist_to_wall = .4975
        ceil_vent_xarea = .1728
        ceil_vent_height = .75
        floor_vent_xarea = .1728
        floor_vent_height = .02
        rel_angle = np.pi/2
        x0 = 0
        y0 = 0
        is_steady = True

        d_imaginary = 2e-4
        ambient = Fluid(species=amb_species, T=amb_temp, P=amb_pressure)
        orifice = Orifice(d_imaginary)
        rho_H2_STP = Fluid(species=rel_species, T = 295, P = 101325).rho

        self.flowrate = Qs1m3 * 60 * 1000

        pressures = []
        for Q in Qs1m3:
            def err_hy3(P):
                g = Fluid(species=rel_species, P=P, T=rel_temp)
                throat = orifice.flow(g, amb_pressure)
                return Q - orifice.mdot(throat)/rho_H2_STP

            pressures.append(fsolve(err_hy3, 10*amb_pressure)[0])

        data = utils.read_csv(os.path.join(DATA_LOC, 'bernardmichelhoussinagbomson-2017-fig9.csv'), header=2)

        # Create concentration data for both release areas, d=4mm and d=27mm
        self.calc_4mm_concentration = []
        self.calc_27mm_concentration = []
        for pressure in pressures:
            fluid = Fluid(species=rel_species, P=pressure, T=rel_temp)
            source = Source(tank_volume, fluid)

            ceil_vent = Vent(ceil_vent_xarea, ceil_vent_height, rel_dis_coeff, 0)
            floor_vent = Vent(floor_vent_xarea, floor_vent_height, rel_dis_coeff, 0)
            enclosure = Enclosure(enclosure_height, floor_ceil_area, rel_height, ceil_vent, floor_vent, dist_to_wall)

            release_4mm = IndoorRelease(source, orifice, ambient, enclosure,
                                        release_area=area_1, theta0=rel_angle, steady=is_steady,
                                        x0=x0, y0=y0, verbose=False)
            release_27mm = IndoorRelease(source, orifice, ambient, enclosure,
                                         release_area=area_2, theta0=rel_angle, steady=is_steady,
                                         x0=x0, y0=y0, verbose=False)

            self.calc_4mm_concentration.append(release_4mm.x_layer[-1] * 100)
            self.calc_27mm_concentration.append(release_27mm.x_layer[-1] * 100)

        self.exp_4mm_concentration = data['y1']
        self.exp_27mm_concentration = data['y2']

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    def test_flowrate_4mm(self):
        title = "Bernard et al. 2017 - Figure 9, Concentration v. Flow Rate, d = 4mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=self.exp_4mm_concentration,
                                calc_vals=self.calc_4mm_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=self.flowrate,
                                exp_y=self.exp_4mm_concentration,
                                calc_y=self.calc_4mm_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Flow Rate (NL/min)',
                                ylabel='Concentration (vol %)')

    def test_flowrate_27mm(self):
        title = "Bernard et al. 2017 - Figure 9, Concentration v. Flow Rate, d = 27mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=self.exp_27mm_concentration,
                                calc_vals=self.calc_27mm_concentration,
                                error_limits=error_limits,
                                units='vol %',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=self.flowrate,
                                exp_y=self.exp_27mm_concentration,
                                calc_y=self.calc_27mm_concentration,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Flow Rate (NL/min)',
                                ylabel='Concentration (vol %)')


if __name__ == "__main__":
    if os.path.exists(OUTPUT_FILE):
        os.remove(OUTPUT_FILE)
    unittest.main()
